/*
  generator -- a program for creating spamdyke test scripts
  Copyright (C) 2015 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * This program recursively explores every possible code path in spamdyke's
 * recipient validation algorithm (see the flowchart in the documentation
 * folder for details).  Because there are hundreds of thousands of possible
 * paths, there's simply no way to create the scripts by hand.  In addition,
 * including all those scripts in the spamdyke tarball would increase its
 * size significantly.
 *
 * Each test is created in its own folder and follows the structure expected
 * by the "run" script.  The script names are generated using the flowchart
 * path they represent.  For example, the name "01y02y05n06y07n..." means
 * the script follows the path Start->1 Yes->2 Yes->5 No->6 Yes->7 No-> in the
 * flowchart. The comments at the top of each script provide more detail about
 * the conditions it creates before qmail is run.
 *
 * Two versions of each script are generated -- one that runs qmail alone and
 * one that runs qmail with spamdyke.  The idea is to test the flowchart's
 * correctness against qmail's actual behavior before testing spamdyke.
 * spamdyke shouldn't reject messages qmail would normally accept and
 * vice-versa.
 *
 * To make the code easier to match to the flowchart, every function in this
 * program begins with a number and two underscores (e.g. f15__).  The numbers
 * are shown on the flowchart in orange circles next to the decision nodes
 * they implement.
 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <math.h>
#include <errno.h>

#define MAX_BUF                         65535
#define STRLEN(X)                       ((int)(sizeof(X) - 1))
#define _STRINGIFY(X)                   #X
#define STRINGIFY(X)                    _STRINGIFY(X)
#define MINVAL(a,b)                     ({ typeof (a) _a = (a); typeof (b) _b = (b); _a < _b ? _a : _b; })
#define MAXVAL(a,b)                     ({ typeof (a) _a = (a); typeof (b) _b = (b); _a > _b ? _a : _b; })
#define PREFIX_SIZE                     1000

#define CLEAN_SCRIPT_INTERVAL           300
/*
 * The letters e and n are deliberately excluded because sometimes it happens
 * they are manipulated into a string that reads "-e" or "-n", which has
 * special meaning when passed to the "echo" command and makes it malfunction.
 * <sigh>
 */
#define RANDOM_ALPHABET                 "abcdfghijklmopqrstuvwxyz0123456789"

#define REJECTION_UNQUALIFIED           0
#define REJECTION_RELAYING              1
#define REJECTION_UNAVAILABLE           2
#define REJECTION_INVALID               3
#define REJECTION_CANNOT_START          4

#define REJECTION_SUFFIX                { "_unqual", "_relay", "_unavail", "_invalid", "_nostart" }
#define REJECTION_DESC                  { "unqualified recipient", "relaying not allowed", "unavailable recipient", "invalid recipient", "qmail cannot start" }

#define ACCEPT_DELIVER                  0
#define ACCEPT_BOUNCE                   1
#define ACCEPT_DISCARD                  2
#define ACCEPT_REMOTE                   3
#define ACCEPT_QUEUE                    4
#define ACCEPT_LOOP                     5

#define ACCEPT_SUFFIX                   { "_deliver", "_bounce", "_discard", "_remote", "_queue", "_loop" }
#define ACCEPT_DESC                     { "local delivery", "bounced to the sender", "discarded", "delivered to a remote address", "stuck in the queue until bounced", "stuck in a delivery loop" }

#define SPAMDYKE_SUFFIX                 "_s"
#define QMAIL_SUFFIX                    "_q"

#define MAX_SUFFIX                      10

#define MAX_FOLDER_LEN                  255

#define STEP(FUNC_NUM,FUNC_NAME,COMMENT,PRE,ACTION,POST) \
  FLOWCHART_STEP(FUNC_NUM,FUNC_NAME,-1,COMMENT,PRE,ACTION,POST,-1,"","",;,"",0)
#define DECISION(FUNC_NUM,FUNC_NAME,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,N_COMMENT,N_PRE,N_ACTION,N_POST) \
  FLOWCHART_STEP(FUNC_NUM,FUNC_NAME,-1,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,-1,N_COMMENT,N_PRE,N_ACTION,N_POST,1)
#define DECISION_LIMITED(FUNC_NUM,FUNC_NAME,MAX_REENTER_Y,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,MAX_REENTER_N,N_COMMENT,N_PRE,N_ACTION,N_POST) \
  FLOWCHART_STEP(FUNC_NUM,FUNC_NAME,MAX_REENTER_Y,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,MAX_REENTER_N,N_COMMENT,N_PRE,N_ACTION,N_POST,1)
#define FLOWCHART_STEP(FUNC_NUM,FUNC_NAME,MAX_REENTER_Y,Y_COMMENT,Y_PRE,Y_ACTION,Y_POST,MAX_REENTER_N,N_COMMENT,N_PRE,N_ACTION,N_POST,ADD_PATH) \
  void FUNC_NAME() \
    { \
    static int reentered_y = 0; \
    static int reentered_n = 0; \
    static char tmp_post[MAX_BUF + 1]; \
    int tmp_strlen_pre; \
    int tmp_strlen_post; \
    int tmp_strlen; \
    if (verbose >= 2) \
      { \
      if (count_only) \
        printf("Counting total scripts %d: %s\n", total_number, STRINGIFY(FUNC_NAME)); \
      else \
        printf("Script %d: %s\n", next_number, STRINGIFY(FUNC_NAME)); \
      fflush(NULL); \
      } \
    if ((MAX_REENTER_N == -1) || \
        (MAX_REENTER_N > reentered_n)) \
      { \
      tmp_strlen_pre = 0; \
      tmp_strlen_post = 0; \
      reentered_n++; \
      if (ADD_PATH) \
        { \
        snprintf(path + strlen_path, MAX_BUF - strlen_path, "%sn", STRINGIFY(FUNC_NUM)); \
        strlen_path += STRLEN(STRINGIFY(FUNC_NUM)) + 1; \
        } \
      snprintf(comments + strlen_comments, MAX_BUF - strlen_comments, "#   %s (%s)\n%n", N_COMMENT, STRINGIFY(FUNC_NAME), &tmp_strlen); \
      strlen_comments += tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      if (STRLEN(N_PRE) > 0) \
        { \
        snprintf(pre + strlen_pre, MAX_BUF - strlen_pre, "\n################################################################################\n# %s (%s)\nLEVEL_%s=%d\n%s%n", N_COMMENT, STRINGIFY(FUNC_NAME), STRINGIFY(FUNC_NAME), reentered_y + reentered_n, randomize(N_PRE, STRLEN(N_PRE)), &tmp_strlen_pre); \
        strlen_pre += tmp_strlen_pre; \
        pre[strlen_pre] = '\0'; \
        } \
      if (STRLEN(N_POST) > 0) \
        { \
        snprintf(tmp_post, MAX_BUF, "################################################################################\n# %s (%s)\nLEVEL_%s=%d\n%s\n%n", N_COMMENT, STRINGIFY(FUNC_NAME), STRINGIFY(FUNC_NAME), reentered_y + reentered_n, randomize(N_POST, STRLEN(N_POST)), &tmp_strlen_post); \
        memmove(post + tmp_strlen_post, post, strlen_post); \
        memcpy(post, tmp_post, tmp_strlen_post); \
        strlen_post += tmp_strlen_post; \
        post[strlen_post] = '\0'; \
        } \
      N_ACTION \
      strlen_comments -= tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      strlen_pre -= tmp_strlen_pre; \
      pre[strlen_pre] = '\0'; \
      if (STRLEN(N_POST) > 0) \
        { \
        strlen_post -= tmp_strlen_post; \
        memmove(post, post + tmp_strlen_post, strlen_post); \
        } \
      post[strlen_post] = '\0'; \
      if (ADD_PATH) \
        { \
        strlen_path -= STRLEN(STRINGIFY(FUNC_NUM)) + 1; \
        path[strlen_path] = '\0'; \
        } \
      reentered_n--; \
      } \
    if ((MAX_REENTER_Y == -1) || \
        (MAX_REENTER_Y > reentered_y)) \
      { \
      tmp_strlen_pre = 0; \
      tmp_strlen_post = 0; \
      reentered_y++; \
      if (ADD_PATH) \
        { \
        snprintf(path + strlen_path, MAX_BUF - strlen_path, "%sy", STRINGIFY(FUNC_NUM)); \
        strlen_path += STRLEN(STRINGIFY(FUNC_NUM)) + 1; \
        } \
      snprintf(comments + strlen_comments, MAX_BUF - strlen_comments, "#   %s (%s)\n%n", Y_COMMENT, STRINGIFY(FUNC_NAME), &tmp_strlen); \
      strlen_comments += tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      if (STRLEN(Y_PRE) > 0) \
        { \
        snprintf(pre + strlen_pre, MAX_BUF - strlen_pre, "\n################################################################################\n# %s (%s)\nLEVEL_%s=%d\n%s%n", Y_COMMENT, STRINGIFY(FUNC_NAME), STRINGIFY(FUNC_NAME), reentered_y + reentered_n, randomize(Y_PRE, STRLEN(Y_PRE)), &tmp_strlen_pre); \
        strlen_pre += tmp_strlen_pre; \
        pre[strlen_pre] = '\0'; \
        } \
      if (STRLEN(Y_POST) > 0) \
        { \
        snprintf(tmp_post, MAX_BUF, "################################################################################\n# %s (%s)\nLEVEL_%s=%d\n%s\n%n", Y_COMMENT, STRINGIFY(FUNC_NAME), STRINGIFY(FUNC_NAME), reentered_y + reentered_n, randomize(Y_POST, STRLEN(Y_POST)), &tmp_strlen_post); \
        memmove(post + tmp_strlen_post, post, strlen_post); \
        memcpy(post, tmp_post, tmp_strlen_post); \
        strlen_post += tmp_strlen_post; \
        post[strlen_post] = '\0'; \
        } \
      Y_ACTION \
      strlen_comments -= tmp_strlen; \
      comments[strlen_comments] = '\0'; \
      strlen_pre -= tmp_strlen_pre; \
      pre[strlen_pre] = '\0'; \
      if (STRLEN(Y_POST) > 0) \
        { \
        strlen_post -= tmp_strlen_post; \
        memmove(post, post + tmp_strlen_post, strlen_post); \
        } \
      post[strlen_post] = '\0'; \
      if (ADD_PATH) \
        { \
        strlen_path -= STRLEN(STRINGIFY(FUNC_NUM)) + 1; \
        path[strlen_path] = '\0'; \
        } \
      reentered_y--; \
      } \
    return; \
    }
#define INC_DEC_FUNC(VARNAME,MAX) \
  "function " VARNAME "_INC()\n" \
  "  {\n" \
  "  " VARNAME "=$[${" VARNAME "}+1]\n" \
  "  if [ ${" VARNAME "} -gt " MAX " ]\n" \
  "  then\n" \
  "    echo \"ERROR: " VARNAME " is greater than " MAX "\"\n" \
  "    outcome=\"failure\"\n" \
  "  fi\n" \
  "  }\n" \
  "function " VARNAME "_DEC()\n" \
  "  {\n" \
  "  " VARNAME "=$[${" VARNAME "}-1]\n" \
  "  if [ ${" VARNAME "} -lt 0 ]\n" \
  "  then\n" \
  "    echo \"ERROR: " VARNAME " is less than 0\"\n" \
  "    outcome=\"failure\"\n" \
  "  fi\n" \
  "  }\n" \

int count_only;
int total_number;
int next_number;
long start_time;
int strlen_comments;
char comments[MAX_BUF + 1];
int strlen_path;
char path[MAX_BUF + 1];
int spamdyke;
char test_name[MAX_BUF + 1];
char test_dir[MAX_BUF + 1];
int strlen_pre;
char pre[MAX_BUF + 1];
int strlen_post;
char post[MAX_BUF + 1];
int max_strlen;
int max_path;
int relayclient_set;
int num_ext_hyphens;
int num_username_hyphens;
int num_username_percents;
int username_truncated;
int using_defaultdelivery;
int dash_emptyable;
int dash_empty;
int envnoathost_exists;
int forwards_only;
int lines_in_file;
int comments_in_file;
int begin_num;
int end_num;
char target_pattern[MAX_BUF + 1];
int strlen_target_pattern;
int verbose;
int invalid_forwards;
int virtualdomains_exact;
int assign_wildcard_domain;
int spamdyke_blind;
int spamdyke_blind_step;
int vdelivermail_loop;

char *alphabet = RANDOM_ALPHABET;
int last_even_letter = -1;
int even_direction = 1;
int last_odd_letter = -1;
int odd_direction = -1;

/*
 * This isn't really very random at all, but a lot of the generated tests will
 * fail if the same random domain names are generated in multiple places within
 * the same script.  Some of the string lengths are so constrained that there
 * are only two random characters in the domain name, so the odds of getting
 * the same two letters in a few scripts out of 100K is high enough that it
 * happens every time.
 *
 * So to try to prevent that, this function does more than just pick purely
 * randomly.  There's randomness involved in picking the start point, but after
 * that it's just sequential iteration.
 */
char *randomize(char *input_text, int strlen_input_text)
  {
  static char return_value[MAX_BUF + 1];
  int i;
  int count;

  if (last_even_letter == -1)
    last_even_letter = random() % STRLEN(RANDOM_ALPHABET);
  if (last_odd_letter == -1)
    {
    last_odd_letter = last_even_letter + (STRLEN(RANDOM_ALPHABET) / 2);
    if (last_odd_letter >= STRLEN(RANDOM_ALPHABET))
      last_odd_letter -= STRLEN(RANDOM_ALPHABET);
    }

  count = 0;
  for (i = 0; i < strlen_input_text; i++)
    if (input_text[i] == '\0')
      {
      if ((count % 2) == 0)
        return_value[i] = alphabet[last_odd_letter];
      else
        {
        return_value[i] = alphabet[last_even_letter];
        last_even_letter += even_direction;
        if ((last_even_letter == STRLEN(RANDOM_ALPHABET)) ||
            (last_even_letter < 0))
          {
          even_direction *= -1;
          last_even_letter += even_direction;

          last_odd_letter += odd_direction;
          if ((last_odd_letter == STRLEN(RANDOM_ALPHABET)) ||
              (last_odd_letter < 0))
            {
            odd_direction *= -1;
            last_odd_letter += odd_direction;
            }
          }
        }

      count++;
      }
    else
      return_value[i] = input_text[i];
  return_value[i] = '\0';

  return(return_value);
  }

FILE *open_file(char *target_path, char *accept_reject, char *suffix)
  {
  static char prefix_format[512] = { '\0' };
  static char filename_format[512] = { '\0' };
  FILE *return_value;
  int tmp_strlen;
  long secs_remaining;
  long secs_elapsed;
  char prefix_dir[MAX_BUF + 1];
  struct stat tmp_stat;

  return_value = NULL;

  if (prefix_format[0] == '\0')
    snprintf(prefix_format, 512, "%%s/%%.%dd", (int)ceil(log10(total_number)));
  if (filename_format[0] == '\0')
    snprintf(filename_format, 512, "%%s/%%.%dd/%%.%dd-%%s%%s%%s%%n", (int)ceil(log10(total_number)), (int)ceil(log10(total_number)));

  snprintf(prefix_dir, MAX_BUF, prefix_format, test_dir, next_number - (next_number % PREFIX_SIZE));
  snprintf(test_name, MAX_BUF, filename_format, test_dir, next_number - (next_number % PREFIX_SIZE), next_number, target_path, accept_reject, suffix, &tmp_strlen);

  printf("Creating test %d of %d: %s", next_number, total_number, test_name);
  if ((next_number < 100) ||
      ((next_number % 20) != 0))
    printf("\n");
  else
    {
    secs_elapsed = time(NULL) - start_time;
    if (secs_elapsed >= 180)
      {
      secs_remaining = (secs_elapsed * (total_number - next_number)) / next_number;
      printf(": \t%ld:%.2ld:%.2ld elapsed, estimated %ld:%.2ld:%.2ld remaining\n", secs_elapsed / 3600, (secs_elapsed % 3600) / 60, (secs_elapsed % 3600) % 60, secs_remaining / 3600, (secs_remaining % 3600) / 60, (secs_remaining % 3600) % 60);
      }
    else
      printf("\n");
    }

  if (stat(prefix_dir, &tmp_stat) == 0)
    {
    if (verbose)
      printf("Found directory %s\n", prefix_dir);

    if ((tmp_stat.st_mode & S_IFDIR) != S_IFDIR)
      {
      fprintf(stderr, "Non-directory already exists where directory should be: %s\n", prefix_dir);
      exit(1);
      }
    }
  else if (errno == ENOENT)
    {
    if (verbose)
      printf("Creating directory %s\n", prefix_dir);

    if (mkdir(prefix_dir, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1)
      {
      fprintf(stderr, "Cannot mkdir %s: %s\n", prefix_dir, strerror(errno));
      exit(1);
      }
    }
  else
    {
    fprintf(stderr, "Cannot stat %s: %s\n", prefix_dir, strerror(errno));
    exit(1);
    }

  if (verbose)
    printf("Creating directory %s\n", test_name);

  if (mkdir(test_name, S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1)
    {
    fprintf(stderr, "Cannot mkdir %s: %s\n", test_name, strerror(errno));
    exit(1);
    }

  snprintf(test_name + tmp_strlen, MAX_BUF - tmp_strlen, "/run.sh");

  if (verbose)
    printf("Creating script %s\n", test_name);

  if ((return_value = fopen(test_name, "w")) == NULL)
    {
    fprintf(stderr, "Cannot fopen %s: %s\n", test_name, strerror(errno));
    exit(1);
    }

  return(return_value);
  }

#define REJECT_TOP \
  "# The message should be rejected: %s\n" \
  "\n" \
  "%s" \
  "\n" \
  "################################################################################\n" \
  "# final steps\n" \
  "if [ -z \"${GREP_TARGET}\" ]\n" \
  "then\n" \
  "  GREP_TARGET=${REJECT_GREP_TARGET}\n" \
  "fi\n"

void reject_message(int rejection_type)
  {
  FILE *script_file;
  char *rejection_suffix[] = REJECTION_SUFFIX;
  char *rejection_desc[] = REJECTION_DESC;

  if (!count_only)
    {
    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        (strlen_target_pattern == 0))
      {
      if (((next_number % CLEAN_SCRIPT_INTERVAL) == 0) &&
          (next_number > 0))
        {
        script_file = open_file("X", "clean_tmp", "");

        fprintf(script_file,
          "# This script was automatically generated by the \"generator\" program.\n"
          "# Any bugs should be fixed there, not here.\n"
          "#\n"
          "# This script deletes all the files from the temporary folder because the\n"
          "# folder will fill up eventually and no more files can be created, which\n"
          "# will cause later scripts to fail.\n"
          "\n"
          "TEST_FOLDER=`echo ${TEST_NUM} | sed -e \"s/_[0-9]*$//\"`\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -print0 2> /dev/null | xargs -0 chmod -R 777\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -print0 2> /dev/null | xargs -0 rm -rf\n"
          "\n"
          "outcome=\"success\"\n"
          );

        fclose(script_file);
        }
      }

    if (((next_number % CLEAN_SCRIPT_INTERVAL) == 0) &&
        (next_number > 0))
      next_number++;

    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        ((strlen_target_pattern == 0) ||
         (strstr(path, target_pattern) != NULL)))
      {
      script_file = open_file(path, rejection_suffix[rejection_type], spamdyke ? SPAMDYKE_SUFFIX : QMAIL_SUFFIX);

      fprintf(script_file,
        "#DEBUG=\"true\"\n"
        "\n"
        "# This script was automatically generated by the \"generator\" program.\n"
        "# Any bugs should be fixed there, not here.\n"
        "#\n"
        "# This script tests recipient validation using the flowchart in the\n"
        "# documentation folder.  It follows this path through the flowchart:\n"
        "%s",
        comments);

      fprintf(script_file,
        randomize(REJECT_TOP, STRLEN(REJECT_TOP)),
        rejection_desc[rejection_type],
        pre);

      fprintf(script_file,
        "\n"
        "################################################################################\n"
        "# check for errors\n"
        "if [ -z \"${outcome}\" ]\n"
        "then\n"
        "\n"
        "################################################################################\n"
        "# empty queue\n"
        "echo Emptying queue and restarting qmail...\n"
        "svc -d /service/qmail-send\n"
        "svc -d /service/qmail-smtpd\n"
        "for i in 2 1 0\n"
        "do\n"
        "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
        "  if [ -z \"${output}\" ]\n"
        "  then\n"
        "    break\n"
        "  else\n"
        "    ${SLEEP_PATH} 1\n"
        "  fi\n"
        "done\n"
        "if [ -n \"${output}\" ]\n"
        "then\n"
        "  svc -k /service/qmail-send\n"
        "  svc -k /service/qmail-smtpd\n"
        "fi\n"
        "rm -f `find /var/qmail/queue/mess /var/qmail/queue/info /var/qmail/queue/local /var/qmail/queue/remote /var/qmail/queue/bounce /var/qmail/queue/intd /var/qmail/queue/todo -type f -print`\n"
        "svc -u /service/qmail-send\n"
        "svc -u /service/qmail-smtpd\n");

      fprintf(script_file,
        "\n"
        "\n"
        "################################################################################\n"
        "# create input file\n"
        "echo \"HELO me\" > ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"MAIL FROM:<${FROM_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"RCPT TO:<${TO_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n");

      if (!spamdyke)
        fprintf(script_file,
          "echo \"DATA\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
          "echo \"To: <${REAL_ADDRESS}>\\r\\nFrom:<${FROM_ADDRESS}>\\r\\nSubject: ${TEST_NUM}\\r\\n\\r\\nFOO!\\r\\n.\\r\\n\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
          );

      fprintf(script_file,
        "echo \"QUIT\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "\n"
        );

      if (spamdyke)
        {
        fprintf(script_file,
          "################################################################################\n"
          "# run diagnostics\n"
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  ${SPAMDYKE_QRV_WITH_VPOPMAIL_PATH} -vv ${TO_DOMAIN} ${TO_USERNAME}\n"
          "fi\n"
          "output=`${TEST_SPAMDYKE_QRV_PATH} -d ${TO_DOMAIN} ${TO_USERNAME} | grep \"QRV-DIAG\" | sed -e \"s/QRV-DIAG: decision path = //\"`\n"
          );
        if (spamdyke_blind)
          fprintf(script_file,
            "expected_output=`echo %s | sed -e \"s/[yn]/_/g\" -e \"s/^\\(0[^7]_\\)*//\" -e \"s/%.2d_.*$/%.2d_/\"`\n",
            path,
            spamdyke_blind_step,
            spamdyke_blind_step
            );
        else
          fprintf(script_file,
            "expected_output=`echo %s | sed -e \"s/[yn]/_/g\" -e \"s/^\\(0[^7]_\\)*//\"`\n",
            path
            );
        fprintf(script_file,
          "if [ -z \"${expected_output}\" ]\n"
          "then\n"
          "  output=\"\"\n"
          "fi\n"
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  echo TO_ADDRESS: ${TO_ADDRESS}\n"
          "  echo REAL_ADDRESS: ${REAL_ADDRESS}\n"
          "  echo VPOPMAIL_ADDRESS: ${VPOPMAIL_ADDRESS}\n"
          "  echo Contents of ${TEMP_FILENAME}:\n"
          "  cat ${TEMP_FILENAME}\n"
          "fi\n"
          "if [ \"${output}\" != \"${expected_output}\" ]\n"
          "then\n"
          "  echo \"ERROR: diagnostic output from spamdyke-qrv differed from expected value:\"\n"
          "  echo \"expected: ${expected_output}\"\n"
          "  echo \"actual:   ${output}\"\n"
          "  outcome=\"failure\"\n"
          "else\n"
          );
        fprintf(script_file,
          "  ################################################################################\n"
          "  # run test\n"
          "  echo \"spamdyke-qrv decision path: ${output}\"\n"
          "  echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${TEST_SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "  if [ -n \"${DEBUG}\" ]\n"
          "  then\n"
          "    ${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -lexcessive --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command \"${TEST_SPAMDYKE_QRV_PATH} -vv\" ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "  else\n"
          "    ${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${TEST_SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "  fi\n"
          "  sync\n"
          "\n"
          );
        switch (rejection_type)
          {
          case REJECTION_UNQUALIFIED:
            fprintf(script_file,
              "  output=`grep \"554 Improper recipient address. Try supplying a domain name.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "  if [ -n \"${output}\" ]\n"
              "  then\n"
              "    output=`grep \"DENIED_UNQUALIFIED_RECIPIENT from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_RELAYING:
            fprintf(script_file,
              "  output=`grep \"554 Refused. Sending to remote addresses (relaying) is not allowed.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "  if [ -n \"${output}\" ]\n"
              "  then\n"
              "    output=`grep \"DENIED_RELAYING from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_UNAVAILABLE:
            fprintf(script_file,
              "  output=`grep \"554 Refused. The recipient is not accepting mail right now.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "  if [ -n \"${output}\" ]\n"
              "  then\n"
              "    output=`grep \"DENIED_UNAVAILABLE_RECIPIENT from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_INVALID:
            fprintf(script_file,
              "  output=`grep \"554 Refused. The recipient address does not exist.\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              "  if [ -n \"${output}\" ]\n"
              "  then\n"
              "    output=`grep \"DENIED_INVALID_RECIPIENT from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
              );
            break;
          case REJECTION_CANNOT_START:
            printf("ERROR: reject_message() was passed REJECTION_CANNOT_START when spamdyke was true.  This should not be possible.\n");
            exit(0);
            break;
          }
        fprintf(script_file,
          "    if [ -n \"${output}\" ]\n"
          "    then\n"
          "      outcome=\"success\"\n"
          "    else\n"
          "      echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "      cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n"
          "fi\n"
          );
        }
      else
        {
        fprintf(script_file,
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  echo TO_ADDRESS: ${TO_ADDRESS}\n"
          "  echo REAL_ADDRESS: ${REAL_ADDRESS}\n"
          "  echo VPOPMAIL_ADDRESS: ${VPOPMAIL_ADDRESS}\n"
          "  echo Contents of ${TEMP_FILENAME}:\n"
          "  cat ${TEMP_FILENAME}\n"
          "fi\n"
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "sync\n"
          "\n"
          );
        if (rejection_type == REJECTION_CANNOT_START)
          fprintf(script_file,
            "output=`grep \"421 unable to read controls (#4.3.0)\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
            "if [ -n \"${output}\" ]\n"
            "then\n"
            "  outcome=\"success\"\n"
            "else\n"
            "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
            "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
            "\n"
            "  outcome=\"failure\"\n"
            "fi\n"
            );
        else if (rejection_type == REJECTION_RELAYING)
          fprintf(script_file,
            "output=`grep \"553 sorry, that domain isn't in my list of allowed rcpthosts (#5.7.1)\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
            "if [ -n \"${output}\" ]\n"
            "then\n"
            "  outcome=\"success\"\n"
            "else\n"
            "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
            "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
            "\n"
            "  outcome=\"failure\"\n"
            "fi\n"
            );
        else
          fprintf(script_file,
            "unset outcome\n"
            "unset MAILDIR\n"
            "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
            "if [ -n \"${output}\" ]\n"
            "then\n"
            "  echo \"Waiting to see if qmail bounces the message...\"\n"
            "\n"
            "  for i in 10 9 8 7 6 5 4 3 2 1 0\n"
            "  do\n"
            "    svc -a /service/qmail-send\n"
            "    svc -a /service/qmail-send\n"
            "    svc -a /service/qmail-send\n"
            "    sync\n"
            "    if [ -f /var/log/qmail/current ]\n"
            "    then\n"
            "      output=`grep -E \"starting delivery [^:]*: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
            "      if [ -n \"${output}\" ]\n"
            "      then\n"
            "        echo \"Found log of bounced message in /var/log/qmail/current\"\n"
            "        echo \"${output}\"\n"
            "        outcome=\"success\"\n"
            "        break\n"
            "      else\n"
            "        output=`grep -E \"starting delivery [^:]*: msg [^ ]* to local\" /var/log/qmail/current`\n"
            "        if [ -n \"${output}\" ]\n"
            "        then\n"
            "          echo \"FOUND LOG OF DELIVERED MESSAGE IN /var/log/qmail/current, SEARCHING TO SEE WHERE IT WENT...\"\n"
            "          output=`grep -rl ${FROM_ADDRESS} ~root ~postmaster ~alias ${VALID_HOMEDIR[0]} ${VALID_HOMEDIR[1]} ${VALID_HOMEDIR[2]} ${VALID_HOMEDIR[3]} ${VALID_HOMEDIR[4]} ${TMPDIR}/${TEST_NUM}-home*`\n"
            "          if [ -n \"${output}\" ]\n"
            "          then\n"
            "            echo \"FOUND MESSAGE IN ${output} - SHOULD HAVE BEEN BOUNCED\"\n"
            "            outcome=\"failure\"\n"
            "            break\n"
            "          else\n"
            "            echo \"CANNOT FIND MESSAGE - MAY HAVE BEEN DISCARDED, SHOULD HAVE BEEN BOUNCED\"\n"
            "          fi\n"
            "        else\n"
            "          output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
            "          if [ -n \"${output}\" ]\n"
            "          then\n"
            "            echo \"Found queued message in /var/qmail/queue/local, waiting to see what happens...\"\n"
            "          else\n"
            "            output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
            "            if [ -n \"${output}\" ]\n"
            "            then\n"
            "              echo \"Found queued message in /var/qmail/queue/remote, waiting to see what happens...\"\n"
            "            fi\n"
            "          fi\n"
            "        fi\n"
            "      fi\n"
            "    fi\n"
            "    ${SLEEP_PATH} 1\n"
            "  done\n"
            "\n"
            "  if [ -z \"${outcome}\" ]\n"
            "  then\n"
            "    echo \"COULD NOT FIND MESSAGE TO ${FROM_ADDRESS} IN LOGS; SHOULD HAVE BEEN BOUNCED\"\n"
            "    outcome=\"failure\"\n"
            "  fi\n"
            "else\n"
            "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
            "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
            "\n"
            "  outcome=\"failure\"\n"
            "fi\n"
            );
        }

      fprintf(script_file, "\nfi\n%s", post);

      fprintf(script_file,
        "\n"
        "echo Restarting qmail...\n"
        "svc -d /service/qmail-send\n"
        "svc -d /service/qmail-smtpd\n"
        "for i in 2 1 0\n"
        "do\n"
        "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
        "  if [ -z \"${output}\" ]\n"
        "  then\n"
        "    break\n"
        "  else\n"
        "    ${SLEEP_PATH} 1\n"
        "  fi\n"
        "done\n"
        "if [ -n \"${output}\" ]\n"
        "then\n"
        "  svc -k /service/qmail-send\n"
        "  svc -k /service/qmail-smtpd\n"
        "fi\n"
        "svc -u /service/qmail-send\n"
        "svc -u /service/qmail-smtpd\n"
        "unset DEBUG\n");

      fprintf(script_file, "\n");

      fclose(script_file);
      }

    next_number++;
    }
  else
    {
    if ((total_number % CLEAN_SCRIPT_INTERVAL) == 0)
      total_number++;

    total_number++;
    max_strlen = MAXVAL(max_strlen, MAXVAL(strlen_pre, strlen_post));
    max_path = MAXVAL(max_path, strlen_path);
    }

  return;
  }

#define ACCEPT_TOP \
  "# The message should be accepted: %s\n" \
  "\n" \
  "%s" \
  "\n" \
  "################################################################################\n" \
  "# final steps\n" \
  "if [ -z \"${GREP_TARGET}\" ]\n" \
  "then\n" \
  "  GREP_TARGET=${REJECT_GREP_TARGET}\n" \
  "fi\n"

void accept_message(int accept_type)
  {
  FILE *script_file;
  char *accept_suffix[] = ACCEPT_SUFFIX;
  char *accept_desc[] = ACCEPT_DESC;

  if (!count_only)
    {
    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        (strlen_target_pattern == 0))
      {
      if (((next_number % CLEAN_SCRIPT_INTERVAL) == 0) &&
          (next_number > 0))
        {
        script_file = open_file("X", "clean_tmp", "");

        fprintf(script_file,
          "# This script was automatically generated by the \"generator\" program.\n"
          "# Any bugs should be fixed there, not here.\n"
          "#\n"
          "# This script deletes all the files from the temporary folder because the\n"
          "# folder will fill up eventually and no more files can be created, which\n"
          "# will cause later scripts to fail.\n"
          "\n"
          "TEST_FOLDER=`echo ${TEST_NUM} | sed -e \"s/_[0-9]*$//\"`\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -print0 2> /dev/null | xargs -0 chmod -R 777\n"
          "find ${TMPDIR} -name ${TEST_FOLDER}* -print0 2> /dev/null | xargs -0 rm -rf\n"
          "\n"
          "outcome=\"success\"\n"
          );

        fclose(script_file);
        }
      }

    if (((next_number % CLEAN_SCRIPT_INTERVAL) == 0) &&
        (next_number > 0))
      next_number++;

    if (((begin_num == -1) ||
         (next_number >= begin_num)) &&
        ((end_num == -1) ||
         (next_number <= end_num)) &&
        ((strlen_target_pattern == 0) ||
         (strstr(path, target_pattern) != NULL)))
      {
      script_file = open_file(path, accept_suffix[accept_type], spamdyke ? SPAMDYKE_SUFFIX : QMAIL_SUFFIX);

      fprintf(script_file,
        "#DEBUG=\"true\"\n"
        "\n"
        "# This script was automatically generated by the \"generator\" program.\n"
        "# Any bugs should be fixed there, not here.\n"
        "#\n"
        "# This script tests recipient validation using the flowchart in the\n"
        "# documentation folder.  It follows this path through the flowchart:\n"
        "%s",
        comments);

      fprintf(script_file,
        randomize(ACCEPT_TOP, STRLEN(ACCEPT_TOP)),
        accept_desc[accept_type],
        pre);

      fprintf(script_file,
        "\n"
        "################################################################################\n"
        "# check for errors\n"
        "if [ -z \"${outcome}\" ]\n"
        "then\n"
        "\n"
        "################################################################################\n"
        "# empty queue\n"
        "echo Emptying queue and restarting qmail...\n"
        "svc -d /service/qmail-send\n"
        "svc -d /service/qmail-smtpd\n"
        "for i in 2 1 0\n"
        "do\n"
        "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
        "  if [ -z \"${output}\" ]\n"
        "  then\n"
        "    break\n"
        "  else\n"
        "    ${SLEEP_PATH} 1\n"
        "  fi\n"
        "done\n"
        "if [ -n \"${output}\" ]\n"
        "then\n"
        "  svc -k /service/qmail-send\n"
        "  svc -k /service/qmail-smtpd\n"
        "fi\n"
        "rm -f `find /var/qmail/queue/mess /var/qmail/queue/info /var/qmail/queue/local /var/qmail/queue/remote /var/qmail/queue/bounce /var/qmail/queue/intd /var/qmail/queue/todo -type f -print`\n"
        "svc -u /service/qmail-send\n"
        "svc -u /service/qmail-smtpd\n");

      fprintf(script_file,
        "\n"
        "\n"
        "################################################################################\n"
        "# create input file\n"
        "echo \"HELO me\" > ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"MAIL FROM:<${FROM_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"RCPT TO:<${TO_ADDRESS}>\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"DATA\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"To: <${REAL_ADDRESS}>\\r\\nFrom: <${FROM_ADDRESS}>\\r\\nSubject: ${TEST_NUM}\\r\\n\\r\\nFOO!\\r\\n.\\r\\n\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "echo \"QUIT\" >> ${TMPDIR}/${TEST_NUM}-input.txt\n"
        "\n"
        );

      if (spamdyke)
        {
        fprintf(script_file,
          "################################################################################\n"
          "# run diagnostics\n"
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  ${SPAMDYKE_QRV_WITH_VPOPMAIL_PATH} -vv ${TO_DOMAIN} ${TO_USERNAME}\n"
          "fi\n"
          "output=`${TEST_SPAMDYKE_QRV_PATH} -d ${TO_DOMAIN} ${TO_USERNAME} | grep \"QRV-DIAG\" | sed -e \"s/QRV-DIAG: decision path = //\"`\n"
          );
        if (spamdyke_blind)
          fprintf(script_file,
            "expected_output=`echo %s | sed -e \"s/[yn]/_/g\" -e \"s/^\\(0[^7]_\\)*//\" -e \"s/%.2d_.*$/%.2d_/\"`\n",
            path,
            spamdyke_blind_step,
            spamdyke_blind_step
            );
        else
          fprintf(script_file,
            "expected_output=`echo %s | sed -e \"s/[yn]/_/g\" -e \"s/^\\(0[^7]_\\)*//\"`\n",
            path
            );
        fprintf(script_file,
          "if [ -z \"${expected_output}\" ]\n"
          "then\n"
          "  output=\"\"\n"
          "fi\n"
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  echo TO_ADDRESS: ${TO_ADDRESS}\n"
          "  echo REAL_ADDRESS: ${REAL_ADDRESS}\n"
          "  echo VPOPMAIL_ADDRESS: ${VPOPMAIL_ADDRESS}\n"
          "  echo Contents of ${TEMP_FILENAME}:\n"
          "  cat ${TEMP_FILENAME}\n"
          "fi\n"
          "if [ \"${output}\" != \"${expected_output}\" ]\n"
          "then\n"
          "  echo \"ERROR: diagnostic output from spamdyke-qrv differed from expected value:\"\n"
          "  echo \"expected: ${expected_output}\"\n"
          "  echo \"actual:   ${output}\"\n"
          "  outcome=\"failure\"\n"
          "else\n"
          );
        fprintf(script_file,
          "################################################################################\n"
          "# run test\n"
          "echo \"spamdyke-qrv decision path: ${output}\"\n"
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${TEST_SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  ${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -lexcessive --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command \"${TEST_SPAMDYKE_QRV_PATH} -vv\" ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "else\n"
          "  ${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${SPAMDYKE_PATH} --log-target stderr -linfo --run-as-user qmaild --reject-recipient invalid --reject-recipient unavailable --recipient-validation-command ${TEST_SPAMDYKE_QRV_PATH} ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "fi\n"
          "sync\n"
          "\n"
          "output=`grep \"ALLOWED from: ${FROM_ADDRESS} to: ${TO_ADDRESS}\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ -n \"${output}\" ]\n"
          "then\n"
          );
        }
      else
        fprintf(script_file,
          "if [ -n \"${DEBUG}\" ]\n"
          "then\n"
          "  echo TO_ADDRESS: ${TO_ADDRESS}\n"
          "  echo REAL_ADDRESS: ${REAL_ADDRESS}\n"
          "  echo VPOPMAIL_ADDRESS: ${VPOPMAIL_ADDRESS}\n"
          "  echo Contents of ${TEMP_FILENAME}:\n"
          "  cat ${TEMP_FILENAME}\n"
          "fi\n"
          "################################################################################\n"
          "# run test\n"
          "echo \"${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\"\n"
          "${SENDRECV_PATH} -t 30 -r 221 -w 0 -- ${QMAIL_CMDLINE} < ${TMPDIR}/${TEST_NUM}-input.txt > ${TMPDIR}/${TEST_NUM}-output.txt 2>&1\n"
          "sync\n"
          "\n"
          );

      if (accept_type == ACCEPT_DELIVER)
        fprintf(script_file,
          "  unset outcome\n"
          "  output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "  if [ -n \"${output}\" ]\n"
          "  then\n"
          "    if [ -n \"${GREP_TARGET}\" ]\n"
          "    then\n"
          "      echo \"Waiting to see if qmail delivers the message...\"\n"
          "\n"
          "      for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "      do\n"
          "        svc -a /service/qmail-send\n"
          "        svc -a /service/qmail-send\n"
          "        svc -a /service/qmail-send\n"
          "        sync\n"
          "        if [ -e \"${GREP_TARGET}\" ]\n"
          "        then\n"
          "          output=`grep -r \"${FROM_ADDRESS}\" ${GREP_TARGET}`\n"
          "        else\n"
          "          output=\"\"\n"
          "        fi\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        fi\n"
          "        ${SLEEP_PATH} 1\n"
          "      done\n"
          "\n"
          "      if [ -z \"${outcome}\" ]\n"
          "      then\n"
          "        output=`grep -E \"starting delivery [^:]*: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          echo \"FOUND LOG OF BOUNCED MESSAGE IN /var/log/qmail/current - SHOULD HAVE BEEN DELIVERED\"\n"
          "          outcome=\"failure\"\n"
          "          break\n"
          "        else\n"
          "          echo \"DID NOT FIND MESSAGE FROM ${FROM_ADDRESS} IN ${GREP_TARGET}; SHOULD HAVE BEEN DELIVERED\"\n"
          "          outcome=\"failure\"\n"
          "        fi\n"
          "      fi\n"
          "    else\n"
          "      echo SCRIPT BUG: GREP_TARGET IS NOT SET!\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n");
      else if (accept_type == ACCEPT_LOOP)
        fprintf(script_file,
          "  unset outcome\n"
          "  output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "  if [ -n \"${output}\" ]\n"
          "  then\n"
          "    echo \"Waiting to see if the message is looping in the queue...\"\n"
          "\n"
          "    for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "    do\n"
          "      svc -a /service/qmail-send\n"
          "      svc -a /service/qmail-send\n"
          "      svc -a /service/qmail-send\n"
          "      sync\n"
          "      output=`grep -E \"failure: message_is_looping_${REAL_ADDRESS}${REAL_RELAYCLIENT}/mail_is_looping/\" /var/log/qmail/current`\n"
          "      if [ -n \"${output}\" ]\n"
          "      then\n"
          "        echo \"Found log of looping message in /var/log/qmail/current\"\n"
          "        outcome=\"success\"\n"
          "        break\n"
          "      else\n"
          "        output=`grep -E \"starting delivery [^:]*: msg [^ ]* to local ${REAL_ADDRESS}\" /var/log/qmail/current | awk '{ print $4 }' | sort | uniq | wc -l`\n"
          "        if [ ${output} -gt 5 ]\n"
          "        then\n"
          "          echo \"Found log of looping message in /var/log/qmail/current\"\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        fi\n"
          "      fi\n"
          "      ${SLEEP_PATH} 1\n"
          "    done\n"
          "\n"
          "    if [ -z \"${outcome}\" ]\n"
          "    then\n"
          "      echo \"DID NOT FIND LOOPING MESSAGE TO ${REAL_ADDRESS} IN /var/log/qmail/current\"\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n");
      else if (accept_type == ACCEPT_BOUNCE)
        fprintf(script_file,
          "unset outcome\n"
          "unset MAILDIR\n"
          "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ -n \"${output}\" ]\n"
          "then\n"
          "  echo \"Waiting to see if qmail bounces the message...\"\n"
          "\n"
          "  for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "  do\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    sync\n"
          "    if [ -f /var/log/qmail/current ]\n"
          "    then\n"
          "      output=`grep -E \"starting delivery [^:]*: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
          "      if [ -n \"${output}\" ]\n"
          "      then\n"
          "        echo \"Found log of bounced message in /var/log/qmail/current\"\n"
          "        outcome=\"success\"\n"
          "        break\n"
          "      else\n"
          "        output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          echo \"Found queued message in /var/qmail/queue/local, continuing to wait...\"\n"
          "        else\n"
          "          output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
          "          if [ -n \"${output}\" ]\n"
          "          then\n"
          "            echo \"Found queued message in /var/qmail/queue/remote, continuing to wait...\"\n"
          "          fi\n"
          "        fi\n"
          "      fi\n"
          "    fi\n"
          "    ${SLEEP_PATH} 1\n"
          "  done\n"
          "\n"
          "  if [ -z \"${outcome}\" ]\n"
          "  then\n"
          "    output=`grep -E \"starting delivery [^:]*: msg [^ ]* to local\" /var/log/qmail/current`\n"
          "    if [ -n \"${output}\" ]\n"
          "    then\n"
          "      echo \"Found log of delivered message in /var/log/qmail/current, searching to see where it went...\"\n"
          "      output=`grep -rl ${FROM_ADDRESS} ~root ~postmaster ~alias ${VALID_HOMEDIR[0]} ${VALID_HOMEDIR[1]} ${VALID_HOMEDIR[2]} ${VALID_HOMEDIR[3]} ${VALID_HOMEDIR[4]} ${TMPDIR}/${TEST_NUM}-home*`\n"
          "      if [ -n \"${output}\" ]\n"
          "      then\n"
          "        echo \"FOUND MESSAGE IN ${output} - SHOULD HAVE BEEN BOUNCED\"\n"
          "        outcome=\"failure\"\n"
          "        break\n"
          "      else\n"
          "        echo \"CANNOT FIND MESSAGE - MAY HAVE BEEN DISCARDED, SHOULD HAVE BEEN BOUNCED\"\n"
          "      fi\n"
          "    else\n"
          "      echo \"COULD NOT FIND LOG OF BOUNCE MESSAGE TO ${FROM_ADDRESS}; SHOULD HAVE BEEN BOUNCED\"\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  fi\n"
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );
      else if (accept_type == ACCEPT_DISCARD)
        fprintf(script_file,
          "unset outcome\n"
          "unset MAILDIR\n"
          "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ -n \"${output}\" ]\n"
          "then\n"
          "  echo \"Waiting to see if qmail discards the message...\"\n"
          "\n"
          "  for i in 10 9 8 7 6 5 4 3 2 1\n"
          "  do\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    sync\n"
          "    if [ -f /var/log/qmail/current ]\n"
          "    then\n"
          "      output=`grep -E \"starting delivery [^:]*: msg [^ ]* to local\" /var/log/qmail/current`\n"
          "      if [ -n \"${output}\" ]\n"
          "      then\n"
          "        echo \"Found log of delivered message in /var/log/qmail/current, searching to see where it went...\"\n"
          "        output=`grep -rl ${FROM_ADDRESS} ~root ~postmaster ~alias ${VALID_HOMEDIR[0]} ${VALID_HOMEDIR[1]} ${VALID_HOMEDIR[2]} ${VALID_HOMEDIR[3]} ${VALID_HOMEDIR[4]} ${TMPDIR}/${TEST_NUM}-home*`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          echo \"FOUND BOUNCED MESSAGE IN ${output} - SHOULD HAVE BEEN DISCARDED\"\n"
          "          outcome=\"failure\"\n"
          "          break\n"
          "        else\n"
          "          echo \"Could not find message - must have been discarded\"\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        fi\n"
          "      else\n"
          "        output=`grep -E \"starting delivery [^:]*: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          echo \"FOUND LOG OF BOUNCED MESSAGE IN /var/log/qmail/current - SHOULD HAVE BEEN DISCARDED\"\n"
          "          echo \"${output}\"\n"
          "          outcome=\"failure\"\n"
          "          break\n"
          "        else\n"
          "          output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
          "          if [ -n \"${output}\" ]\n"
          "          then\n"
          "            echo \"Found queued message in /var/qmail/queue/local, waiting to see what happens...\"\n"
          "          else\n"
          "            output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
          "            if [ -n \"${output}\" ]\n"
          "            then\n"
          "              echo \"FOUND QUEUED MESSAGE IN /var/qmail/queue/remote - SHOULD HAVE BEEN DISCARDED\"\n"
          "              echo \"${output}\"\n"
          "              outcome=\"failure\"\n"
          "              break\n"
          "            fi\n"
          "          fi\n"
          "        fi\n"
          "      fi\n"
          "    fi\n"
          "    ${SLEEP_PATH} 1\n"
          "  done\n"
          "\n"
          "  if [ -z \"${outcome}\" ]\n"
          "  then\n"
          "    echo \"Could not find message - must have been discarded\"\n"
          "    outcome=\"success\"\n"
          "  fi\n"
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );
      else if (accept_type == ACCEPT_REMOTE)
        fprintf(script_file,
          "  unset outcome\n"
          "  output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "  if [ -n \"${output}\" ]\n"
          "  then\n"
          "    echo \"Waiting to see if qmail delivers the message...\"\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "\n"
          "    for i in 10 9 8 7 6 5 4 3 2 1 0\n"
          "    do\n"
          "      sync\n"
          "      for i in 0 1 2 3 4\n"
          "      do\n"
          "        output=`grep \" to remote ${VALID_ADDRESS[$i]}\" /var/log/qmail/current`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        else\n"
          "          output=`grep \" to remote ${INVALID_ADDRESS[$i]}\" /var/log/qmail/current`\n"
          "          if [ -n \"${output}\" ]\n"
          "          then\n"
          "            outcome=\"success\"\n"
          "            break\n"
          "          fi\n"
          "        fi\n"
          "      done\n"
          "      if [ -z \"${outcome}\" ]\n"
          "      then\n"
          "        output=`grep -r \"${REAL_ADDRESS}\" /var/qmail/queue/remote/*`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        else\n"
          "          output=`grep -r \"${FROM_ADDRESS}\" /var/qmail/queue/remote/*`\n"
          "          if [ -n \"${output}\" ]\n"
          "          then\n"
          "            outcome=\"success\"\n"
          "            break\n"
          "          fi\n"
          "        fi\n"
          "      else\n"
          "        break\n"
          "      fi\n"
          "      ${SLEEP_PATH} 1\n"
          "    done\n"
          "\n"
          "    if [ -z \"${outcome}\" ]\n"
          "    then\n"
          "      echo \"COULD NOT FIND MESSAGE TO ${REAL_ADDRESS} OR FROM ${FROM_ADDRESS} IN /var/qmail/queue/remote OR TO ${VALID_ADDRESS[0]} OR TO ${VALID_ADDRESS[1]} OR TO ${VALID_ADDRESS[2]} OR TO ${VALID_ADDRESS[3]} OR TO ${VALID_ADDRESS[4]} OR TO ${INVALID_ADDRESS[0]} OR TO ${INVALID_ADDRESS[1]} OR TO ${INVALID_ADDRESS[2]} OR TO ${INVALID_ADDRESS[3]} OR TO ${INVALID_ADDRESS[4]} IN /var/log/qmail/current; SHOULD HAVE BEEN DELIVERED REMOTELY\"\n"
          "      outcome=\"failure\"\n"
          "    fi\n"
          "  else\n"
          "    echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "    cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "    outcome=\"failure\"\n"
          "  fi\n");
      else if (accept_type == ACCEPT_QUEUE)
        fprintf(script_file,
          "unset outcome\n"
          "unset MAILDIR\n"
          "output=`grep -E \"250 ok [0-9]* qp [0-9]*\" ${TMPDIR}/${TEST_NUM}-output.txt`\n"
          "if [ -n \"${output}\" ]\n"
          "then\n"
          "  echo \"Waiting to see if qmail queues the message...\"\n"
          "\n"
          "  queue_loop=0\n"
          "  for i in 10 9 8 7 6 5 4 3 2 1\n"
          "  do\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    svc -a /service/qmail-send\n"
          "    sync\n"
          "    if [ -f /var/log/qmail/current ]\n"
          "    then\n"
          "      output=`grep -r \"T${REAL_ADDRESS}\" /var/qmail/queue/local`\n"
          "      if [ -n \"${output}\" ]\n"
          "      then\n"
          "        if [ ${queue_loop} -gt 3 ]\n"
          "        then\n"
          "          echo \"Message appears to be stuck in the queue, as expected\"\n"
          "          outcome=\"success\"\n"
          "          break\n"
          "        else\n"
          "          echo \"Found queued message in /var/qmail/queue/local, waiting to see if it stays there...\"\n"
          "          queue_loop=$[${queue_loop}+1]\n"
          "        fi\n"
          "      else\n"
          "        output=`grep -E \"starting delivery [^:]*: msg [^ ]* to local\" /var/log/qmail/current`\n"
          "        if [ -n \"${output}\" ]\n"
          "        then\n"
          "          echo \"Found log of delivered message in /var/log/qmail/current, searching to see where it went...\"\n"
          "          output=`grep -rl ${FROM_ADDRESS} ~root ~postmaster ~alias ${VALID_HOMEDIR[0]} ${VALID_HOMEDIR[1]} ${VALID_HOMEDIR[2]} ${VALID_HOMEDIR[3]} ${VALID_HOMEDIR[4]} ${TMPDIR}/${TEST_NUM}-home*`\n"
          "          if [ -n \"${output}\" ]\n"
          "          then\n"
          "            echo \"FOUND BOUNCED MESSAGE IN ${output} - SHOULD HAVE BEEN QUEUED\"\n"
          "            outcome=\"failure\"\n"
          "            break\n"
          "          else\n"
          "            echo \"COULD NOT FIND MESSAGE - MAY HAVE BEEN DISCARDED, SHOULD HAVE BEEN QUEUED\"\n"
          "            outcome=\"failure\"\n"
          "            break\n"
          "          fi\n"
          "        else\n"
          "          output=`grep -E \"starting delivery [^:]*: msg [^ ]* to remote ${FROM_ADDRESS}\" /var/log/qmail/current`\n"
          "          if [ -n \"${output}\" ]\n"
          "          then\n"
          "            echo \"FOUND LOG OF BOUNCED MESSAGE IN /var/log/qmail/current - SHOULD BE QUEUED FOR LOCAL DELIVERY\"\n"
          "            echo \"${output}\"\n"
          "            outcome=\"failure\"\n"
          "            break\n"
          "          else\n"
          "            output=`grep -r \"T${FROM_ADDRESS}\" /var/qmail/queue/remote`\n"
          "            if [ -n \"${output}\" ]\n"
          "            then\n"
          "              echo \"FOUND QUEUED MESSAGE IN /var/qmail/queue/remote - SHOULD BE QUEUED FOR LOCAL DELIVERY\"\n"
          "              echo \"${output}\"\n"
          "              outcome=\"failure\"\n"
          "              break\n"
          "            fi\n"
          "          fi\n"
          "        fi\n"
          "      fi\n"
          "    fi\n"
          "    ${SLEEP_PATH} 1\n"
          "  done\n"
          "\n"
          "  if [ -z \"${outcome}\" ]\n"
          "  then\n"
          "    echo \"TIME EXPIRED BEFORE A CONCLUSION COULD BE REACHED, MESSAGE SHOULD HAVE BEEN QUEUED\"\n"
          "    outcome=\"failure\"\n"
          "  fi\n"
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          );

      if (spamdyke)
        fprintf(script_file,
          "else\n"
          "  echo OUTPUT IN tmp/${TEST_NUM}-output.txt:\n"
          "  cat ${TMPDIR}/${TEST_NUM}-output.txt\n"
          "\n"
          "  outcome=\"failure\"\n"
          "fi\n"
          "fi\n"
          );

      fprintf(script_file, "\nfi\n%s", post);

      fprintf(script_file,
        "\n"
        "echo Restarting qmail...\n"
        "svc -d /service/qmail-send\n"
        "svc -d /service/qmail-smtpd\n"
        "for i in 2 1 0\n"
        "do\n"
        "  output=`ps ax | grep qmail-s[em][nt][dp] | sed -e \"/^.*grep.*$/d\" -e \"/^.*tail.*$/d\" -e \"/^.*supervise.*$/d\" -e \"/^.*tcpserver.*$/d\"`\n"
        "  if [ -z \"${output}\" ]\n"
        "  then\n"
        "    break\n"
        "  else\n"
        "    ${SLEEP_PATH} 1\n"
        "  fi\n"
        "done\n"
        "if [ -n \"${output}\" ]\n"
        "then\n"
        "  svc -k /service/qmail-send\n"
        "  svc -k /service/qmail-smtpd\n"
        "fi\n"
        "svc -u /service/qmail-send\n"
        "svc -u /service/qmail-smtpd\n"
        "unset DEBUG\n");

      fprintf(script_file, "\n");

      fclose(script_file);
      }

    next_number++;
    }
  else
    {
    if ((total_number % CLEAN_SCRIPT_INTERVAL) == 0)
      total_number++;

    total_number++;
    max_strlen = MAXVAL(max_strlen, MAXVAL(strlen_pre, strlen_post));
    max_path = MAXVAL(max_path, strlen_path);
    }

  return;
  }


void f54__more_valias_destinations();
void f39__more_lines_in_file();
void f31__does_passwd_contain_temp_username();
void f26__does_wildcard_filename_exist();
void f07__does_recipient_contain_percent();


DECISION(68, f68__is_forward_address_valid,
  "local forward address is valid",
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"&${VALID_ADDRESS[$CURRENT_VALID]}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  /* Only possible to reach this point because spamdyke-qrv does not have
     vpopmail support and the domain isn't in the assign file. */
  if (spamdyke_blind &&
      (invalid_forwards > 0))
    accept_message(ACCEPT_BOUNCE);
  else
    accept_message(ACCEPT_DELIVER);
  lines_in_file--;,
  "",
  "forward address is not valid",
  "REJECT_GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"&${INVALID_ADDRESS[$CURRENT_VALID]}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  invalid_forwards++;
  f39__more_lines_in_file();
  invalid_forwards--;
  lines_in_file--;,
  "")


DECISION(67, f67__is_forward_domain_in_locals,
  "forward address domain is in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f67_locals.bu.${LEVEL_f67__is_forward_domain_in_locals}\n"
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f67_locals.${LEVEL_f67__is_forward_domain_in_locals}\n"
  "echo ${VALID_DOMAIN[$CURRENT_VALID]} >> ${TMPDIR}/${TEST_NUM}-f67_locals.${LEVEL_f67__is_forward_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f67_locals.${LEVEL_f67__is_forward_domain_in_locals} /var/qmail/control/locals\n",
  f68__is_forward_address_valid();,
  "cp ${TMPDIR}/${TEST_NUM}-f67_locals.bu.${LEVEL_f67__is_forward_domain_in_locals} /var/qmail/control/locals\n",
  "forward address domain is not in locals",
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"&${VALID_ADDRESS[$CURRENT_VALID]}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  accept_message(ACCEPT_REMOTE);
  lines_in_file--;,
  "")


DECISION_LIMITED(66, f66__is_forward_address_same_as_current,
  1,
  "forward address is the same as the current address",
  "CURRENT_VALID_INC\n"
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"&${REAL_ADDRESS}${REAL_RELAYCLIENT}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  invalid_forwards++;
  f39__more_lines_in_file();
  invalid_forwards--;
  lines_in_file--;,
  "",
  -1,
  "forward address is not the same as the current address",
  "CURRENT_VALID_INC\n",
  f67__is_forward_domain_in_locals();,
  "")


DECISION(65, f65__does_maildir_exist,
  "maildir exists and is writable",
  "VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]=${TMPDIR}/${TEST_NUM}-f65_home.${LEVEL_f65__does_maildir_exist}\n"
  "mkdir -p ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/Maildir/new ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/Maildir/cur ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/Maildir/tmp\n"
  "chmod -R 700 ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/Maildir\n"
  "chown -R $6 ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}\n"
  "chgrp -R $7 ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/\" >> ${TEMP_FILENAME}\n"
  "GREP_TARGET=${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/Maildir/\n",
  lines_in_file++;
  accept_message(ACCEPT_DELIVER);
  lines_in_file--;,
  "",
  "maildir does not exist",
  "VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]=${TMPDIR}/${TEST_NUM}-f65_home.${LEVEL_f65__does_maildir_exist}\n"
  "REJECT_GREP_TARGET=${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/Maildir/\n"
  "rm -rf ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" ${VDELIVERMAIL_HOMEDIR[${LEVEL_f65__does_maildir_exist}]}/\" >> ${TEMP_FILENAME}\n",
  if (spamdyke &&
      !spamdyke_blind)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_BOUNCE);,
  "")


DECISION(64, f64__is_forward_address_valid,
  "local forward address is valid",
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" ${VALID_ADDRESS[$CURRENT_VALID]}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  accept_message(ACCEPT_DELIVER);
  lines_in_file--;,
  "",
  "forward address is not valid",
  "REJECT_GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" ${INVALID_ADDRESS[$CURRENT_VALID]}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  invalid_forwards++;
  f39__more_lines_in_file();
  invalid_forwards--;
  lines_in_file--;,
  "")


DECISION(63, f63__is_forward_domain_in_locals,
  "forward address domain is in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f63_locals.bu.${LEVEL_f63__is_forward_domain_in_locals}\n"
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f63_locals.${LEVEL_f63__is_forward_domain_in_locals}\n"
  "echo ${VALID_DOMAIN[$CURRENT_VALID]} >> ${TMPDIR}/${TEST_NUM}-f63_locals.${LEVEL_f63__is_forward_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f63_locals.${LEVEL_f63__is_forward_domain_in_locals} /var/qmail/control/locals\n",
  f64__is_forward_address_valid();,
  "cp ${TMPDIR}/${TEST_NUM}-f63_locals.bu.${LEVEL_f63__is_forward_domain_in_locals} /var/qmail/control/locals\n",
  "forward address domain is not in locals",
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" ${VALID_ADDRESS[$CURRENT_VALID]}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  accept_message(ACCEPT_REMOTE);
  lines_in_file--;,
  "")


DECISION_LIMITED(62, f62__is_forward_address_same_as_current,
  1,
  "forward address is the same as the current address",
  "CURRENT_VALID_INC\n"
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" ${REAL_ADDRESS}${REAL_RELAYCLIENT}\" >> ${TEMP_FILENAME}\n",
  if (spamdyke &&
      !spamdyke_blind)
    reject_message(REJECTION_INVALID);
  else
    accept_message(ACCEPT_BOUNCE);,
  "",
  -1,
  "forward address is not the same as the current address",
  "CURRENT_VALID_INC\n",
  f63__is_forward_domain_in_locals();,
  "")


DECISION(61, f61__does_last_parameter_begin_with_slash,
  "last parameter to vdelivermail begins with a slash",
  "",
  f65__does_maildir_exist();,
  "",
  "last parameter to vdelivermail does not begin with a slash",
  "",
  f62__is_forward_address_same_as_current();,
  "")


DECISION(60, f60__is_last_parameter_delete,
  "last parameter to vdelivermail is \"delete\"",
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" delete\" >> ${TEMP_FILENAME}\n"
  "GREP_TARGET=\"${REJECT_GREP_TARGET} ${HOMEDIR}/${TO_USERNAME}/Maildir/\"\n",
  lines_in_file++;
  accept_message(ACCEPT_DISCARD);
  lines_in_file--;,
  "",
  "last parameter to vdelivermail is not \"delete\"",
  "",
  f61__does_last_parameter_begin_with_slash();,
  "")


DECISION(59, f59__is_last_parameter_bounce_no_mailbox,
  "last parameter to vdelivermail is \"bounce-no-mailbox\"",
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n",
  if (spamdyke &&
      !spamdyke_blind)
    reject_message(REJECTION_INVALID);
  else
    accept_message(ACCEPT_BOUNCE);,
  "",
  "last parameter to vdelivermail is not \"bounce-no-mailbox\"",
  "",
  f60__is_last_parameter_delete();,
  "")


DECISION(58, f58__recipient_found_by_vuserinfo,
  "recipient address is found by vuserinfo",
  "${VPOPMAIL_VADDUSER_PATH} -q NOQUOTA ${VPOPMAIL_ADDRESS} $3\n"
  "GREP_TARGET=${HOMEDIR}/${VPOPMAIL_USERNAME}/Maildir/\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  if (reentered_n == 0)
    accept_message(ACCEPT_DELIVER);
  lines_in_file--;,
  "",
  "recipient address is not found by vuserinfo",
  "",
  if (reentered_y == 0)
    f59__is_last_parameter_bounce_no_mailbox();,
  "")


DECISION(57, f57__is_valias_destination_valid,
  "valias destination is valid",
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n"
  "${VPOPMAIL_VALIAS_PATH} -i ${VALID_ADDRESS[$CURRENT_VALID]} ${VPOPMAIL_ADDRESS}\n",
  lines_in_file++;
  if (reentered_n == 0)
    accept_message(ACCEPT_DELIVER);
  lines_in_file--;,
  "",
  "valias destination is not valid",
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n"
  "${VPOPMAIL_VALIAS_PATH} -i ${INVALID_ADDRESS[$CURRENT_VALID]} ${VPOPMAIL_ADDRESS}\n",
  lines_in_file++;
  invalid_forwards++;
  if (reentered_y == 0)
    f54__more_valias_destinations();
  invalid_forwards--;
  lines_in_file--;,
  "")


DECISION(56, f56__is_valias_domain_in_locals,
  "valias destination domain name is in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f56_locals.bu.${LEVEL_f56__is_valias_domain_in_locals}\n"
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f56_locals.${LEVEL_f56__is_valias_domain_in_locals}\n"
  "echo ${VALID_DOMAIN[$CURRENT_VALID]} >> ${TMPDIR}/${TEST_NUM}-f56_locals.${LEVEL_f56__is_valias_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f56_locals.${LEVEL_f56__is_valias_domain_in_locals} /var/qmail/control/locals\n",
  f57__is_valias_destination_valid();,
  "cp ${TMPDIR}/${TEST_NUM}-f56_locals.bu.${LEVEL_f56__is_valias_domain_in_locals} /var/qmail/control/locals\n",
  "valias destination domain name is not in locals",
  "GREP_TARGET=${VALID_GREP_TARGET[$CURRENT_VALID]}\n"
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n"
  "${VPOPMAIL_VALIAS_PATH} -i ${VALID_ADDRESS[$CURRENT_VALID]} ${VPOPMAIL_ADDRESS}\n",
  lines_in_file++;
  accept_message(ACCEPT_REMOTE);
  lines_in_file--;,
  "")


DECISION_LIMITED(55, f55__is_valias_destination_same_as_current,
  1,
  "valias destination is the same as the current address",
  "CURRENT_VALID_INC\n"
  "${VPOPMAIL_VALIAS_PATH} -i ${VPOPMAIL_ADDRESS} ${VPOPMAIL_ADDRESS}\n",
  invalid_forwards++;
  f54__more_valias_destinations();
  invalid_forwards--;,
  "",
  -1,
  "valias destination is not the same as the current address",
  "CURRENT_VALID_INC\n",
  f56__is_valias_domain_in_locals();,
  "")


DECISION_LIMITED(54, f54__more_valias_destinations,
  2,
  "recipient address has more valias destinations",
  "",
  f55__is_valias_destination_same_as_current();,
  "",
  -1,
  "recipient address has no more valias destinations",
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  if (reentered_y > 0)
    {
    if (spamdyke &&
        !spamdyke_blind)
      reject_message(REJECTION_INVALID);
    else
      accept_message(ACCEPT_BOUNCE);
    }
  lines_in_file--;,
  "")


DECISION(53, f53__does_recipient_have_valias,
  "recipient has valias destinations",
  "",
  if (reentered_n == 0)
    f54__more_valias_destinations();,
  "",
  "recipient has no valias destinations",
  "",
  f58__recipient_found_by_vuserinfo();,
  "")


DECISION(52, f52__does_spamdyke_qrv_have_vpopmail_support,
  "spamdyke-qrv has vpopmail support",
  "",
  f53__does_recipient_have_valias();,
  "",
  "spamdyke-qrv does not have vpopmail support",
  "TEST_SPAMDYKE_QRV_PATH=${SPAMDYKE_QRV_NO_VPOPMAIL_PATH}\n",
  spamdyke_blind++;
  spamdyke_blind_step = 52;
  invalid_forwards++;
  f53__does_recipient_have_valias();
  invalid_forwards--;
  spamdyke_blind--;,
  "")


DECISION(51, f51__is_address_configured_for_vpopmail,
  "the virtualdomains and assign files are correct for vpopmail and the username is valid",
  "",
  if (virtualdomains_exact &&
      assign_wildcard_domain &&
      (num_username_percents == 0))
    {
    if (spamdyke)
      f52__does_spamdyke_qrv_have_vpopmail_support();
    else
      f53__does_recipient_have_valias();
    },
  "",
  "either the virtualdomains and/or assign files are incorrect for vpopmail or the username is not valid",
  "echo \"|${VPOPMAIL_VDELIVERMAIL_PATH} \\\"\\\" bounce-no-mailbox\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  if (!virtualdomains_exact ||
      !assign_wildcard_domain ||
      (num_username_percents > 0))
    {
    if (spamdyke)
      reject_message(REJECTION_INVALID);
    else
      accept_message(ACCEPT_BOUNCE);
    }
  lines_in_file--;,
  "")


DECISION(50, f50__is_program_vdelivermail_with_two_parameters,
  "program is vdelivermail with two parameters",
  "cat ${TEMP_FILENAME} | sed -e \"/^.*f47_tofile.*$/d\" > ${TMPDIR}/${TEST_NUM}-f50_tempfile.${LEVEL_f50__is_program_vdelivermail_with_two_parameters}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f50_tempfile.${LEVEL_f50__is_program_vdelivermail_with_two_parameters} ${TEMP_FILENAME}\n",
  lines_in_file--;
  f51__is_address_configured_for_vpopmail();
  lines_in_file++;,
  "",
  "program is not vdelivermail with two parameters",
  "",
  accept_message(ACCEPT_DELIVER);,
  "")


DECISION(49, f49__does_program_exist,
  "program exists",
  "GREP_TARGET=${HOMEDIR}/mail_output\n"
  "cp ${TOFILE_PATH} ${TMPDIR}/${TEST_NUM}-f47_tofile.${LEVEL_f47__is_line_program}\n"
  "chmod 755 ${TMPDIR}/${TEST_NUM}-f47_tofile.${LEVEL_f47__is_line_program}\n",
  f50__is_program_vdelivermail_with_two_parameters();,
  "",
  "program does not exist",
  "REJECT_GREP_TARGET=${HOMEDIR}/mail_output\n"
  "rm -f ${TMPDIR}/${TEST_NUM}-f47_tofile.${LEVEL_f47__is_line_program}\n",
  if (spamdyke &&
      !spamdyke_blind)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "")


DECISION(48, f48__is_program_allowed,
  "program deliveries are allowed",
  "",
  if (!forwards_only)
    f49__does_program_exist();,
  "",
  "program deliveries are not allowed",
  "",
  if (forwards_only)
    {
    if (spamdyke &&
        !spamdyke_blind)
      reject_message(REJECTION_UNAVAILABLE);
    else
      accept_message(ACCEPT_QUEUE);
    },
  "")


DECISION(47, f47__is_line_program,
  "line in file is a program",
  "echo \"|${TMPDIR}/${TEST_NUM}-f47_tofile.${LEVEL_f47__is_line_program} -f ${HOMEDIR}/mail_output -t 60\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f48__is_program_allowed();
  lines_in_file--;,
  "",
  "line in file is not a program, it must be a forward address",
  "",
  f66__is_forward_address_same_as_current();,
  "")


/*
 * Creating the Maildir here should reference LEVEL_f44__is_line_maildir,
 * since step 44 puts the path to the folder in the .qmail file.
 */
DECISION(46, f46__does_maildir_exist,
  "maildir exists",
  "GREP_TARGET=${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}/new\n"
  "mkdir -p ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}\n"
  "mkdir ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}/cur\n"
  "mkdir ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}/new\n"
  "mkdir ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}/tmp\n"
  "sync\n"
  "chmod -R 700 ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}\n"
  "chown -R $6 ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}\n"
  "chgrp -R $7 ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}\n",
  accept_message(ACCEPT_DELIVER);,
  "",
  "maildir does not exist",
  "REJECT_GREP_TARGET=${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}/new\n"
  "rm -rf ${HOMEDIR}/maildir.${LEVEL_f44__is_line_maildir}\n",
  if (spamdyke &&
      !spamdyke_blind)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "")


DECISION(45, f45__is_maildir_allowed,
  "maildir deliveries are allowed",
  "",
  if (!forwards_only)
    f46__does_maildir_exist();,
  "",
  "maildir deliveries are not allowed",
  "",
  if (forwards_only)
    {
    if (spamdyke &&
        !spamdyke_blind)
      reject_message(REJECTION_UNAVAILABLE);
    else
      accept_message(ACCEPT_QUEUE);
    },
  "")


DECISION(44, f44__is_line_maildir,
  "line in file is a maildir",
  "echo \"./maildir.${LEVEL_f44__is_line_maildir}/\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f45__is_maildir_allowed();
  lines_in_file--;,
  "",
  "line in file is not a maildir",
  "",
  f47__is_line_program();,
  "")


/*
 * Creating the mbox here should reference LEVEL_f41__is_line_mbox,
 * since step 41 puts the path to the mbox in the .qmail file.
 */
DECISION(43, f43__is_mbox_writeable,
  "mbox is writeable",
  "GREP_TARGET=${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "touch ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "sync\n"
  "chmod 600 ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "chown $6 ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "chgrp $7 ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n",
  accept_message(ACCEPT_DELIVER);,
  "",
  "mbox is not writeable",
  "REJECT_GREP_TARGET=${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "touch ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "sync\n"
  "chown root ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n"
  "chmod 600 ${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\n",
  if (spamdyke &&
      !spamdyke_blind)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "")


DECISION(42, f42__is_mbox_allowed,
  "mbox deliveries are allowed",
  "",
  if (!forwards_only)
    f43__is_mbox_writeable();,
  "",
  "mbox deliveries are not allowed",
  "",
  if (forwards_only)
    {
    if (spamdyke &&
        !spamdyke_blind)
      reject_message(REJECTION_UNAVAILABLE);
    else
      accept_message(ACCEPT_QUEUE);
    },
  "")


DECISION(41, f41__is_line_mbox,
  "line in file is an mbox",
  "echo \"${HOMEDIR}/mbox.${LEVEL_f41__is_line_mbox}\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  f42__is_mbox_allowed();
  lines_in_file--;,
  "",
  "line in file is not an mbox",
  "",
  f44__is_line_maildir();,
  "")


DECISION(40, f40__is_line_comment,
  "line in file is a comment",
  "REJECT_GREP_TARGET=${HOMEDIR}/`cat /var/qmail/control/defaultdelivery`\n"
  "echo \"# This line is a meaningless comment.\" >> ${TEMP_FILENAME}\n",
  lines_in_file++;
  comments_in_file++;
  if (!using_defaultdelivery)
    f39__more_lines_in_file();
  comments_in_file--;
  lines_in_file--;,
  "",
  "line in file is not a comment",
  "",
  f41__is_line_mbox();,
  "")


DECISION_LIMITED(39, f39__more_lines_in_file,
  2,
  "more lines in file",
  "",
  if (!using_defaultdelivery ||
      (reentered_y == 1))
    f40__is_line_comment();,
  "",
  -1,
  "no more lines in file",
  "",
  if (lines_in_file > 0)
    {
    if (spamdyke &&
        !spamdyke_blind)
      reject_message(REJECTION_INVALID);
    else if (vdelivermail_loop > 0)
      accept_message(ACCEPT_LOOP);
    else if (invalid_forwards > 0)
      accept_message(ACCEPT_BOUNCE);
    else
      accept_message(ACCEPT_DISCARD);
    }
  else if (reentered_y > 0)
    {
    if (spamdyke &&
        !spamdyke_blind)
      reject_message(REJECTION_INVALID);
    else
      accept_message(ACCEPT_DISCARD);
    },
  "")


DECISION(38, f38__is_first_line_blank,
  "first line is blank",
  "echo \"\" > ${TEMP_FILENAME}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "",
  "first line is not blank",
  "",
  f39__more_lines_in_file();,
  "")


DECISION(37, f37__is_file_owner_executable,
  "file is owner-executable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_EXECUTABLE_MODE[${LEVEL_f37__is_file_owner_executable}]=${st_mode}\n"
  "chmod u+x ${TEMP_FILENAME}\n",
  forwards_only = 1;
  f38__is_first_line_blank();
  forwards_only = 0;,
  "chmod ${FILE_EXECUTABLE_MODE[${LEVEL_f37__is_file_owner_executable}]} ${TEMP_FILENAME}\n",
  "file is not owner-executable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_EXECUTABLE_MODE[${LEVEL_f37__is_file_owner_executable}]=${st_mode}\n"
  "chmod u-x ${TEMP_FILENAME}\n",
  f38__is_first_line_blank();,
  "chmod ${FILE_EXECUTABLE_MODE[${LEVEL_f37__is_file_owner_executable}]} ${TEMP_FILENAME}\n")


DECISION(36, f36__is_file_empty,
  "file is empty",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_EMPTY_MODE[${LEVEL_f36__is_file_empty}]=${st_mode}\n"
  "rm -f ${TEMP_FILENAME}\n"
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n"
  "chmod ${FILE_EMPTY_MODE[${LEVEL_f36__is_file_empty}]} ${TEMP_FILENAME}\n"
  "FILE_EMPTY_TEMP_FILENAME[${LEVEL_f36__is_file_empty}]=${TEMP_FILENAME}\n"
  "TEMP_FILENAME=/var/qmail/control/defaultdelivery\n"
  "if [ -f /var/qmail/control/defaultdelivery ]\n"
  "then\n"
  "  cp /var/qmail/control/defaultdelivery ${TMPDIR}/${TEST_NUM}-f36_defaultdelivery.bu.${LEVEL_f36__is_file_empty}\n"
  "  rm -f /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "touch /var/qmail/control/defaultdelivery\n",
  using_defaultdelivery = 1;
  f39__more_lines_in_file();
  using_defaultdelivery = 0;,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f36_defaultdelivery.bu.${LEVEL_f36__is_file_empty} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f36_defaultdelivery.bu.${LEVEL_f36__is_file_empty} /var/qmail/control/defaultdelivery\n"
  "else\n"
  "  rm -f /var/qmail/control/defaultdelivery\n"
  "fi\n",
  "file is not empty",
  "",
  f37__is_file_owner_executable();,
  "")


DECISION(35, f35__is_file_world_writable,
  "file is world-writeable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_WRITABLE_MODE[${LEVEL_f35__is_file_world_writable}]=${st_mode}\n"
  "chmod o+w ${TEMP_FILENAME}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "chmod ${FILE_WRITABLE_MODE[${LEVEL_f35__is_file_world_writable}]} ${TEMP_FILENAME}\n",
  "file is not world-writeable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_WRITABLE_MODE[${LEVEL_f35__is_file_world_writable}]=${st_mode}\n"
  "chmod o-w ${TEMP_FILENAME}\n",
  f36__is_file_empty();,
  "chmod ${FILE_WRITABLE_MODE[${LEVEL_f35__is_file_world_writable}]} ${TEMP_FILENAME}\n")


DECISION(34, f34__is_file_readable,
  "file is readable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_READABLE_MODE[${LEVEL_f34_is_file_readable}]=${st_mode}\n"
  "chmod o+r ${TEMP_FILENAME}\n",
  f35__is_file_world_writable();,
  "chmod ${FILE_READABLE_MODE[${LEVEL_f34_is_file_readable}]} ${TEMP_FILENAME}\n",
  "file is not readable",
  "eval `${STMODE_PATH} ${TEMP_FILENAME}`\n"
  "FILE_READABLE_MODE[${LEVEL_f34_is_file_readable}]=${st_mode}\n"
  "chown root ${TEMP_FILENAME}\n"
  "chmod 600 ${TEMP_FILENAME}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "chmod ${FILE_READABLE_MODE[${LEVEL_f34_is_file_readable}]} ${TEMP_FILENAME}\n")


DECISION(33, f33__does_passwd_contain_alias,
  "passwd contains alias",
  "HOMEDIR=\"${TMPDIR}/${TEST_NUM}-home.${LEVEL_f33__does_passwd_contain_alias}\"\n"
  "DASH=\"-\"\n"
  "EXT=`echo ${REAL_USERNAME} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "VPOPMAIL_USERNAME=${EXT//:/.}${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-${#VIRTUALDOMAINS_USERNAME_ADDITION}]}\n"
  "VPOPMAIL_ADDRESS=${VPOPMAIL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f34_master.passwd.bu.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cat /etc/master.passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f34_master.passwd.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  echo \"alias:*************:$6:$7::0:0:alias:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f34_master.passwd.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_master.passwd.${LEVEL_f33__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f34_passwd.bu.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cat /etc/passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f34_passwd.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  echo \"alias:x:$6:$7:alias:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f34_passwd.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_passwd.${LEVEL_f33__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n"
  "sync\n"
  "mkdir -p ${HOMEDIR}\n"
  "echo \"${HOMEDIR}/mbox\" > ${HOMEDIR}/.qmail-valid-user-default\n"
  "touch ${HOMEDIR}/mbox\n"
  "chown -R $6 ${HOMEDIR}\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  dash_emptyable = 0;
  dash_empty = 0;
  f26__does_wildcard_filename_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_master.passwd.bu.${LEVEL_f33__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_passwd.bu.${LEVEL_f33__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n",
  "passwd does not contain alias",
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f34_master.passwd.bu.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cat /etc/master.passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f34_master.passwd.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_master.passwd.${LEVEL_f33__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f34_passwd.bu.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cat /etc/passwd | sed -e \"/^alias:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f34_passwd.${LEVEL_f33__does_passwd_contain_alias}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_passwd.${LEVEL_f33__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n",
  if (spamdyke)
    reject_message(REJECTION_INVALID);
  else
    accept_message(ACCEPT_QUEUE);,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_master.passwd.bu.${LEVEL_f33__does_passwd_contain_alias} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f34_passwd.bu.${LEVEL_f33__does_passwd_contain_alias} /etc/passwd\n"
  "fi\n")


DECISION_LIMITED(32, f32__does_temp_username_contain_hyphen,
  2,
  "temp username contains hyphen",
  "USERNAME_HYPHEN_TEMP_USERNAME[${LEVEL_f32__does_temp_username_contain_hyphen}]=${TEMP_USERNAME}\n"
  "TEMP_USERNAME=`echo ${TEMP_USERNAME} | sed -e \"s/-[^-]*$//\"`\n",
  if (num_username_hyphens > 0)
    {
    num_username_hyphens--;
    num_ext_hyphens++;
    username_truncated++;
    f31__does_passwd_contain_temp_username();
    username_truncated--;
    num_ext_hyphens--;
    num_username_hyphens++;
    },
  "",
  -1,
  "temp username does not contain a hyphen",
  "USERNAME_HYPHEN_REAL_USERNAME[${LEVEL_f32__does_temp_username_contain_hyphen}]=${REAL_USERNAME}\n"
  "USERNAME_HYPHEN_REAL_ADDRESS[${LEVEL_f32__does_temp_username_contain_hyphen}]=${REAL_ADDRESS}\n"
  "USERNAME_HYPHEN_TO_USERNAME[${LEVEL_f32__does_temp_username_contain_hyphen}]=${TO_USERNAME}\n"
  "USERNAME_HYPHEN_TO_ADDRESS[${LEVEL_f32__does_temp_username_contain_hyphen}]=${TO_ADDRESS}\n"
  "USERNAME_HYPHEN_TEMP_USERNAME[${LEVEL_f32__does_temp_username_contain_hyphen}]=${TEMP_USERNAME}\n"
  "TRUNCATED=${REAL_USERNAME:${#TEMP_USERNAME}:$[${#REAL_USERNAME}-${#TEMP_USERNAME}]}\n"
  "if [ -n \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}`echo ${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#TRUNCATED}]]} | sed -e \"s/-/_/g\"`${TRUNCATED}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=`echo ${TO_USERNAME:0:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-${#TRUNCATED}]} | sed -e \"s/-/_/g\"`${TO_USERNAME:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-${#TRUNCATED}]:${#TRUNCATED}}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}`echo ${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#TRUNCATED}]]} | sed -e \"s/-/_/g\"`${TRUNCATED}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=`echo ${TO_USERNAME:0:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-$[${#TRUNCATED}-${#REAL_RELAYCLIENT}]]} | sed -e \"s/-/_/g\"`${TO_USERNAME:$[$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]-$[${#TRUNCATED}-${#REAL_RELAYCLIENT}]]:$[${#TRUNCATED}-${#REAL_RELAYCLIENT}]}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "TEMP_USERNAME=`echo ${TEMP_USERNAME} | sed -e \"s/-/_/g\"`\n",
  if (num_username_hyphens == 0)
    f33__does_passwd_contain_alias();,
  "")


DECISION(31, f31__does_passwd_contain_temp_username,
  "passwd contains temp username",
  "HOMEDIR=\"${TMPDIR}/${TEST_NUM}-home.${LEVEL_f31__does_passwd_contain_temp_username}\"\n"
  "if [ \"${TEMP_USERNAME}\" == \"${REAL_USERNAME}\" ]\n"
  "then\n"
  "  DASH=\"\"\n"
  "else\n"
  "  DASH=\"-\"\n"
  "fi\n"
  "EXT=`echo ${REAL_USERNAME:${#TEMP_USERNAME}:$[${#REAL_USERNAME}-${#TEMP_USERNAME}]} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\" -e \"s/^-//\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "VPOPMAIL_USERNAME=${EXT//:/.}${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-${#VIRTUALDOMAINS_USERNAME_ADDITION}]}\n"
  "VPOPMAIL_ADDRESS=${VPOPMAIL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f31_master.passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cat /etc/master.passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f31_master.passwd.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  echo \"${TEMP_USERNAME}:*************:$6:$7::0:0:${TEMP_USERNAME}:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f31_master.passwd.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_master.passwd.${LEVEL_f31__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f31_passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cat /etc/passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f31_passwd.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  echo \"${TEMP_USERNAME}:x:$6:$7:${TEMP_USERNAME}:${HOMEDIR}:/bin/bash\" >> ${TMPDIR}/${TEST_NUM}-f31_passwd.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_passwd.${LEVEL_f31__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n"
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chown $6 ${HOMEDIR}\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  dash_emptyable = 0;
  dash_empty = (username_truncated) ? 0 : 1;
  f26__does_wildcard_filename_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_master.passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n",
  "passwd does not contain temp username",
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp /etc/master.passwd ${TMPDIR}/${TEST_NUM}-f31_master.passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cat /etc/master.passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f31_master.passwd.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_master.passwd.${LEVEL_f31__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp /etc/passwd ${TMPDIR}/${TEST_NUM}-f31_passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cat /etc/passwd | sed -e \"/^${TEMP_USERNAME}:.*$/d\" > ${TMPDIR}/${TEST_NUM}-f31_passwd.${LEVEL_f31__does_passwd_contain_temp_username}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_passwd.${LEVEL_f31__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n",
  f32__does_temp_username_contain_hyphen();,
  "if [ -f /etc/master.passwd ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_master.passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username} /etc/master.passwd\n"
  "  pwd_mkdb /etc/master.passwd\n"
  "else\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f31_passwd.bu.${LEVEL_f31__does_passwd_contain_temp_username} /etc/passwd\n"
  "fi\n")


DECISION(30, f30__is_temp_username_over_31_chars,
  "temp username is over 31 characters",
  "OVER32_REAL_USERNAME[${LEVEL_f30__is_temp_username_over_31_chars}]=${REAL_USERNAME}\n"
  "OVER32_TO_USERNAME[${LEVEL_f30__is_temp_username_over_31_chars}]=${TO_USERNAME}\n"
  "WILDCARD_DOT=\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\n"
  "WILDCARD_DASH=\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}-${WILDCARD_DOT}.${WILDCARD_DASH}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ -n \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME}-${WILDCARD_DOT}.${WILDCARD_DASH}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}-${WILDCARD_DOT}.${WILDCARD_DASH}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "TEMP_USERNAME=${REAL_USERNAME:0:32}\n",
  num_username_hyphens++;
  username_truncated++;
  f32__does_temp_username_contain_hyphen();
  username_truncated--;
  num_username_hyphens--;,
  "",
  "temp username is not over 31 characters",
  "OVER32_REAL_USERNAME[${LEVEL_f30__is_temp_username_over_31_chars}]=${REAL_USERNAME}\n"
  "OVER32_TO_USERNAME[${LEVEL_f30__is_temp_username_over_31_chars}]=${TO_USERNAME}\n"
  "WILDCARD_DOT=\0\n"
  "WILDCARD_DASH=\0\n"
  "WILDCARD_DASH2=\0\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#REAL_RELAYCLIENT}]]}\n"
  "  if [ ${#REAL_USERNAME} -gt $[25-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#REAL_RELAYCLIENT}]] ]\n"
  "  then\n"
  "    STRLEN_REAL_USERNAME=$[25-$[${#VIRTUALDOMAINS_USERNAME_ADDITION}+${#REAL_RELAYCLIENT}]]\n"
  "  else\n"
  "    STRLEN_REAL_USERNAME=${#REAL_USERNAME}\n"
  "  fi\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-${#VIRTUALDOMAINS_USERNAME_ADDITION}]}\n"
  "  if [ ${#REAL_USERNAME} -gt $[25-${#VIRTUALDOMAINS_USERNAME_ADDITION}] ]\n"
  "  then\n"
  "    STRLEN_REAL_USERNAME=$[25-${#VIRTUALDOMAINS_USERNAME_ADDITION}]\n"
  "  else\n"
  "    STRLEN_REAL_USERNAME=${#REAL_USERNAME}\n"
  "  fi\n"
  "fi\n"
  "REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}`echo ${REAL_USERNAME:0:${STRLEN_REAL_USERNAME}} | sed -e \"s/-*$//\"`.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}\n"
  "TO_USERNAME=`echo ${TO_USERNAME:0:${STRLEN_REAL_USERNAME}} | sed -e \"s/-*$//\"`.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ -n \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "  REAL_USERNAME=${REAL_USERNAME}${REAL_RELAYCLIENT}\n"
  "fi\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TEMP_USERNAME=${REAL_USERNAME}\n",
  num_username_hyphens += 2;
  f31__does_passwd_contain_temp_username();
  num_username_hyphens -= 2;,
  "")


DECISION(29, f29__is_wildcard_dash_empty,
  "dash field is empty",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f29_assign.bu.${LEVEL_f29__is_wildcard_dash_empty}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f29_assign.bu.${LEVEL_f29__is_wildcard_dash_empty}`\n"
  "do\n"
  "  if [ \"${line:0:1}\" == \"=\" ]\n"
  "  then\n"
  "    pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "    if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "    then\n"
  "      echo ${line} >> ${TMPDIR}/${TEST_NUM}-f29_assign.${LEVEL_f29__is_wildcard_dash_empty}\n"
  "    fi\n"
  "  else\n"
  "    if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "      pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "      if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "      then\n"
  "        if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "        then\n"
  "          echo ${line} >> ${TMPDIR}/${TEST_NUM}-f29_assign.${LEVEL_f29__is_wildcard_dash_empty}\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "DASH=\"\"\n"
  "TEMP_FILENAME=/var/qmail/control/defaultdelivery\n"
  "if [ -f /var/qmail/control/defaultdelivery ]\n"
  "then\n"
  "  cp /var/qmail/control/defaultdelivery ${TMPDIR}/${TEST_NUM}-f30_defaultdelivery.bu.${LEVEL_f29__is_wildcard_dash_empty}\n"
  "  rm -f /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "touch /var/qmail/control/defaultdelivery\n"
  "echo \"+${PATTERN}:${USER}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f29_assign.${LEVEL_f29__is_wildcard_dash_empty}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f29_assign.${LEVEL_f29__is_wildcard_dash_empty}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f29_assign.${LEVEL_f29__is_wildcard_dash_empty} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  if (dash_emptyable)
    {
    using_defaultdelivery = 1;
    f39__more_lines_in_file();
    using_defaultdelivery = 0;
    },
  "rm -f /var/qmail/control/defaultdelivery\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f30_defaultdelivery.bu.${LEVEL_f29__is_wildcard_dash_empty} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f30_defaultdelivery.bu.${LEVEL_f29__is_wildcard_dash_empty} /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "cp ${TMPDIR}/${TEST_NUM}-f29_assign.bu.${LEVEL_f29__is_wildcard_dash_empty} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "dash field is not empty",
  "",
  if (!dash_empty)
    reject_message(REJECTION_INVALID);,
  "")


DECISION(28, f28__is_ext_empty,
  "ext is empty",
  "",
  if (num_ext_hyphens < 0)
    f29__is_wildcard_dash_empty();,
  "",
  "ext is not empty",
  "EXT_EMPTY_EXT[${LEVEL_f28__is_ext_empty}]=${EXT}\n"
  "EXT_EMPTY_TEMP_FILENAME[${LEVEL_f28__is_ext_empty}]=${TEMP_FILENAME}\n"
  "EXT=\"\"\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}default\n",
  if (num_ext_hyphens == 0)
    {
    num_ext_hyphens--;
    f26__does_wildcard_filename_exist();
    num_ext_hyphens++;
    },
  "")


DECISION_LIMITED(27, f27__does_ext_contain_hyphen,
  2,
  "ext contains hyphen",
  "EXT_HYPHEN_EXT[${LEVEL_f27__does_ext_contain_hyphen}]=${EXT}\n"
  "EXT_HYPHEN_TEMP_FILENAME[${LEVEL_f27__does_ext_contain_hyphen}]=${TEMP_FILENAME}\n"
  "EXT=`echo ${EXT} | sed -e \"s/-[^-]*$//\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}-default\n",
  if (num_ext_hyphens > 0)
    {
    num_ext_hyphens--;
    f26__does_wildcard_filename_exist();
    num_ext_hyphens++;
    },
  "",
  -1,
  "ext does not contain a hyphen",
  "",
  if (num_ext_hyphens <= 0)
    f28__is_ext_empty();,
  "")


DECISION(26, f26__does_wildcard_filename_exist,
  "constructed filename exists",
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n",
  f34__is_file_readable();,
  "",
  "constructed filename does not exist",
  "rm -f ${TEMP_FILENAME}\n",
  f27__does_ext_contain_hyphen();,
  "")


DECISION(25, f25__is_wildcard_home_directory_sticky,
  "home directory is sticky",
  "chmod +t ${HOMEDIR}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "chmod -t ${HOMEDIR}\n",
  "home directory is not sticky",
  "chmod -t ${HOMEDIR}\n",
  f26__does_wildcard_filename_exist();,
  "")


DECISION(24, f24__is_wildcard_home_directory_world_writable,
  "home directory is world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f24__is_wildcard_home_directory_world_writable}]=${st_mode}\n"
  "chmod o+w ${HOMEDIR}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "chmod ${WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f24__is_wildcard_home_directory_world_writable}]} ${HOMEDIR}\n",
  "home directory is not world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f24__is_wildcard_home_directory_world_writable}]=${st_mode}\n"
  "chmod o-w ${HOMEDIR}\n",
  f25__is_wildcard_home_directory_sticky();,
  "chmod ${WILDCARD_HOMEDIR_WRITABLE_MODE[${LEVEL_f24__is_wildcard_home_directory_world_writable}]} ${HOMEDIR}\n")


DECISION(23, f23__is_wildcard_home_directory_readable,
  "home directory is readable",
  "chmod u+r ${HOMEDIR}\n",
  f24__is_wildcard_home_directory_world_writable();,
  "",
  "home directory is not readable",
  "chmod u-r,g-r,o-r ${HOMEDIR}\n",
  reject_message(REJECTION_INVALID);,
  "")


DECISION(22, f22__does_wildcard_home_directory_exist,
  "home directory exists",
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chmod 700 ${HOMEDIR}\n"
  "chown -R $6 ${HOMEDIR}\n"
  "chgrp -R $7 ${HOMEDIR}\n",
  f23__is_wildcard_home_directory_readable();,
  "",
  "home directory does not exist",
  "rm -rf ${HOMEDIR}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "")


DECISION(21, f21__is_exact_match_dash_empty,
  "dash field is empty",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f21_assign.bu.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f21_assign.bu.${LEVEL_f21__is_exact_match_dash_empty} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f21_assign.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "for line in `cat /var/qmail/users/assign`\n"
  "do\n"
  "  if [ \"${line:0:1}\" == \"=\" ]\n"
  "  then\n"
  "    pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "    if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "    then\n"
  "      echo ${line} >> ${TMPDIR}/${TEST_NUM}-f21_assign.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "    fi\n"
  "  else\n"
  "    if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "      pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "      if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "      then\n"
  "        if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "        then\n"
  "          echo ${line} >> ${TMPDIR}/${TEST_NUM}-f21_assign.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "DASH=\"\"\n"
  "TEMP_FILENAME=/var/qmail/control/defaultdelivery\n"
  "if [ -f /var/qmail/control/defaultdelivery ]\n"
  "then\n"
  "  cp /var/qmail/control/defaultdelivery ${TMPDIR}/${TEST_NUM}-f21_defaultdelivery.bu.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "  rm /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "touch /var/qmail/control/defaultdelivery\n"
  "echo \"=${PATTERN}:${REAL_USERNAME}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f21_assign.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f21_assign.${LEVEL_f21__is_exact_match_dash_empty}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f21_assign.${LEVEL_f21__is_exact_match_dash_empty} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  using_defaultdelivery = 1;
  f39__more_lines_in_file();
  using_defaultdelivery = 0;,
  "cp ${TMPDIR}/${TEST_NUM}-f21_assign.bu.${LEVEL_f21__is_exact_match_dash_empty} /var/qmail/users/assign\n"
  "rm /var/qmail/control/defaultdelivery\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f21_defaultdelivery.bu.${LEVEL_f21__is_exact_match_dash_empty} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f21_defaultdelivery.bu.${LEVEL_f21__is_exact_match_dash_empty} /var/qmail/control/defaultdelivery\n"
  "fi\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "dash field is not empty",
  "",
  if (!dash_empty)
    reject_message(REJECTION_INVALID);,
  "")


DECISION(20, f20__does_exact_match_filename_exist_default,
  "constructed filename exists",
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n",
  f34__is_file_readable();,
  "",
  "constructed filename does not exist",
  "rm -f ${TEMP_FILENAME}\n",
  f21__is_exact_match_dash_empty();,
  "")


DECISION(19, f19__does_exact_match_filename_exist,
  "constructed filename exists",
  "touch ${TEMP_FILENAME}\n"
  "sync\n"
  "chown $6 ${TEMP_FILENAME}\n"
  "chgrp $7 ${TEMP_FILENAME}\n",
  f34__is_file_readable();,
  "",
  "constructed filename does not exist",
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}default\n",
  f20__does_exact_match_filename_exist_default();,
  "")


DECISION(18, f18__is_exact_match_home_directory_sticky,
  "home directory is sticky",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f18__is_exact_match_home_directory_sticky}]=${st_mode}\n"
  "chmod +t ${HOMEDIR}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "chmod ${EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f18__is_exact_match_home_directory_sticky}]} ${HOMEDIR}\n",
  "home directory is not sticky",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f18__is_exact_match_home_directory_sticky}]=${st_mode}\n"
  "chmod -t ${HOMEDIR}\n",
  f19__does_exact_match_filename_exist();,
  "chmod ${EXACT_HOMEDIR_STICKY_MODE[${LEVEL_f18__is_exact_match_home_directory_sticky}]} ${HOMEDIR}\n")


DECISION(17, f17__is_exact_match_home_directory_world_writable,
  "home directory is world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f17__is_exact_match_home_directory_world_writable}]=${st_mode}\n"
  "chmod o+w ${HOMEDIR}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "chmod ${EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f17__is_exact_match_home_directory_world_writable}]} ${HOMEDIR}\n",
  "home directory is not world-writable",
  "eval `${STMODE_PATH} ${HOMEDIR}`\n"
  "EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f17__is_exact_match_home_directory_world_writable}]=${st_mode}\n"
  "chmod o-w ${HOMEDIR}\n",
  f18__is_exact_match_home_directory_sticky();,
  "chmod ${EXACT_HOMEDIR_WRITABLE_MODE[${LEVEL_f17__is_exact_match_home_directory_world_writable}]} ${HOMEDIR}\n")


DECISION(16, f16__is_exact_match_home_directory_readable,
  "home directory is readable",
  "chmod u+r ${HOMEDIR}\n",
  f17__is_exact_match_home_directory_world_writable();,
  "",
  "home directory is not readable",
  "chmod u-r,g-r,o-r ${HOMEDIR}\n",
  reject_message(REJECTION_INVALID);,
  "")


DECISION(15, f15__does_exact_match_home_directory_exist,
  "home directory exists",
  "mkdir -p ${HOMEDIR}\n"
  "sync\n"
  "chmod 700 ${HOMEDIR}\n"
  "chown -R $6 ${HOMEDIR}\n"
  "chgrp -R $7 ${HOMEDIR}\n",
  f16__is_exact_match_home_directory_readable();,
  "",
  "home directory does not exist",
  "rm -rf ${HOMEDIR}\n",
  if (spamdyke)
    reject_message(REJECTION_UNAVAILABLE);
  else
    accept_message(ACCEPT_QUEUE);,
  "")


DECISION(14, f14__is_wildcard_match_domain_and_hyphen,
  "wildcard match is exactly the domain name and a hyphen",
  "WILDCARD_REAL_USERNAME[${LEVEL_f14__is_wildcard_match_domain_and_hyphen}]=${REAL_USERNAME}\n"
  "WILDCARD_TO_USERNAME[${LEVEL_f14__is_wildcard_match_domain_and_hyphen}]=${TO_USERNAME}\n"
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f14_assign.bu.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f14_assign.bu.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}`\n"
  "do\n"
  "  if [ \"${line}\" != \".\" ]\n"
  "  then\n"
  "    if [ \"${line:0:1}\" == \"=\" ]\n"
  "    then\n"
  "      pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "      if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "      then\n"
  "        echo ${line} >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "      fi\n"
  "    else\n"
  "      if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "        pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "        if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "        then\n"
  "          if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "          then\n"
  "            echo ${line} >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "          fi\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "PATTERN=`echo ${REAL_USERNAME} | sed -e \"s/-[^-]*//g\"`-\n"
  "WILDCARD_ADDITION=\0\0\0\0\0\n"
  "USER=${REAL_DOMAIN}\n"
  "HOMEDIR=${TMPDIR}/${TEST_NUM}-home.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}.\0\0\0\0\0\n"
  "DASH=-DASH${LEVEL_f14__is_wildcard_match_domain_and_hyphen}.\0\0\0\0\0-\n"
  "FILE_EXT=EXT${LEVEL_f14__is_wildcard_match_domain_and_hyphen}.\0\0\0\0\0-\0\0\0\0\0\n"
  "EXT=`echo ${FILE_EXT} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  EXT=${EXT}${REAL_USERNAME:${#PATTERN}:$[${#REAL_USERNAME}-$[${#PATTERN}+${#REAL_RELAYCLIENT}]]}:${WILDCARD_ADDITION}${REAL_RELAYCLIENT}\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}.${WILDCARD_ADDITION}${REAL_RELAYCLIENT}\n"
  "else\n"
  "  EXT=${EXT}${REAL_USERNAME:${#PATTERN}:$[${#REAL_USERNAME}-${#PATTERN}]}:${WILDCARD_ADDITION}\n"
  "  REAL_USERNAME=${REAL_USERNAME}.${WILDCARD_ADDITION}\n"
  "fi\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}.${WILDCARD_ADDITION}${PERCENTHACK_USERNAME_ADDITION}\n"
  "VPOPMAIL_USERNAME=${EXT//:/.}\n"
  "VPOPMAIL_ADDRESS=${VPOPMAIL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -n \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "echo \"+${PATTERN}:${USER}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  num_ext_hyphens = num_username_hyphens;
  dash_emptyable = 1;
  dash_empty = 0;
  assign_wildcard_domain++;
  f22__does_wildcard_home_directory_exist();
  assign_wildcard_domain--;
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "cp ${TMPDIR}/${TEST_NUM}-f14_assign.bu.${LEVEL_f14__is_wildcard_match_domain_and_hyphen} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "wildcard match is not exactly the domain name and a hyphen",
  "WILDCARD_REAL_USERNAME[${LEVEL_f14__is_wildcard_match_domain_and_hyphen}]=${REAL_USERNAME}\n"
  "WILDCARD_TO_USERNAME[${LEVEL_f14__is_wildcard_match_domain_and_hyphen}]=${TO_USERNAME}\n"
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f14_assign.bu.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f14_assign.bu.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}`\n"
  "do\n"
  "  if [ \"${line}\" != \".\" ]\n"
  "  then\n"
  "    if [ \"${line:0:1}\" == \"=\" ]\n"
  "    then\n"
  "      pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "      if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "      then\n"
  "        echo ${line} >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "      fi\n"
  "    else\n"
  "      if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "        pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "        if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "        then\n"
  "          if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "          then\n"
  "            echo ${line} >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "          fi\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "PATTERN=`echo ${REAL_USERNAME} | sed -e \"s/-[^-]*$/-/\"`\n"
  "WILDCARD_ADDITION=\0\0\0\0\0\n"
  "USER=${REAL_DOMAIN}\n"
  "HOMEDIR=${TMPDIR}/${TEST_NUM}-home.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}.\0\0\0\0\0\n"
  "DASH=-DASH${LEVEL_f14__is_wildcard_match_domain_and_hyphen}.\0\0\0\0\0-\n"
  "FILE_EXT=EXT${LEVEL_f14__is_wildcard_match_domain_and_hyphen}.\0\0\0\0\0-\0\0\0\0\0\n"
  "EXT=`echo ${FILE_EXT} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  EXT=${EXT}${REAL_USERNAME:${#PATTERN}:$[${#REAL_USERNAME}-$[${#PATTERN}+${#REAL_RELAYCLIENT}]]}:${WILDCARD_ADDITION}${REAL_RELAYCLIENT}\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}.${WILDCARD_ADDITION}${REAL_RELAYCLIENT}\n"
  "else\n"
  "  EXT=${EXT}${REAL_USERNAME:${#PATTERN}:$[${#REAL_USERNAME}-${#PATTERN}]}:${WILDCARD_ADDITION}\n"
  "  REAL_USERNAME=${REAL_USERNAME}.${WILDCARD_ADDITION}\n"
  "fi\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}.${WILDCARD_ADDITION}${PERCENTHACK_USERNAME_ADDITION}\n"
  "VPOPMAIL_USERNAME=${EXT//:/.}\n"
  "VPOPMAIL_ADDRESS=${VPOPMAIL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -n \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "echo \"+${PATTERN}:${USER}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f14_assign.${LEVEL_f14__is_wildcard_match_domain_and_hyphen} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  num_ext_hyphens = 1;
  dash_emptyable = 1;
  dash_empty = 0;
  f22__does_wildcard_home_directory_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "cp ${TMPDIR}/${TEST_NUM}-f14_assign.bu.${LEVEL_f14__is_wildcard_match_domain_and_hyphen} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "")


DECISION(13, f13__does_cdb_contain_wildcard_match,
  "cdb contains recipient wildcard match",
  "",
  f14__is_wildcard_match_domain_and_hyphen();,
  "",
  "cdb does not contain recipient wildcard match",
  "WILDCARD_REAL_USERNAME[${LEVEL_f13__does_cdb_contain_wildcard_match}]=${REAL_USERNAME}\n"
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f13_assign.bu.${LEVEL_f13__does_cdb_contain_wildcard_match}\n"
  "cat /dev/null > /var/qmail/users/assign\n"
  "for line in `cat ${TMPDIR}/${TEST_NUM}-f13_assign.bu.${LEVEL_f13__does_cdb_contain_wildcard_match}`\n"
  "do\n"
  "  if [ \"${line}\" != \".\" ]\n"
  "  then\n"
  "    if [ \"${line:0:1}\" == \"=\" ]\n"
  "    then\n"
  "      pattern=`echo ${line} | sed -e \"s/^=//\" -e \"s/:.*//\"`\n"
  "      if [ \"${REAL_USERNAME}\" != \"${pattern}\" ]\n"
  "      then\n"
  "        echo ${line} >> ${TMPDIR}/${TEST_NUM}-f13_assign.${LEVEL_f13__does_cdb_contain_wildcard_match}\n"
  "      fi\n"
  "    else\n"
  "      if [ \"${line:0:1}\" == \"+\" ]\n"
  "      then\n"
  "        pattern=`echo ${line} | sed -e \"s/^+//\" -e \"s/:.*//\"`\n"
  "        if [ ${#REAL_USERNAME} -ge ${#pattern} ]\n"
  "        then\n"
  "          if [ \"${pattern}\" != \"${REAL_USERNAME:0:${#pattern}}\" ]\n"
  "          then\n"
  "            echo ${line} >> ${TMPDIR}/${TEST_NUM}-f13_assign.${LEVEL_f13__does_cdb_contain_wildcard_match}\n"
  "          fi\n"
  "        fi\n"
  "      fi\n"
  "    fi\n"
  "  fi\n"
  "done\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f13_assign.${LEVEL_f13__does_cdb_contain_wildcard_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f13_assign.${LEVEL_f13__does_cdb_contain_wildcard_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "WILDCARD_DOT=\0\n"
  "WILDCARD_DASH=\0\n"
  "WILDCARD_DASH2=\0\n"
  "TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ -n \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME}.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}.${WILDCARD_DOT}-${WILDCARD_DASH}-${WILDCARD_DASH2}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "fi\n"
  "TEMP_USERNAME=${REAL_USERNAME}\n",
  num_username_hyphens += 2;
  num_ext_hyphens = -1;
  f30__is_temp_username_over_31_chars();
  num_username_hyphens -= 2;,
  "cp ${TMPDIR}/${TEST_NUM}-f13_assign.bu.${LEVEL_f13__does_cdb_contain_wildcard_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "")

DECISION(12, f12__does_cdb_contain_exact_match,
  "cdb contains recipient exact match",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_exact_match} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "PATTERN=${REAL_USERNAME}\n"
  "USER=${REAL_USERNAME}\n"
  "HOMEDIR=${TMPDIR}/${TEST_NUM}-home.${LEVEL_f12__does_cdb_contain_exact_match}.\0\0\0\0\0\n"
  "DASH=-DASH${LEVEL_f12__does_cdb_contain_exact_match}.\0\0\0\0\0-\n"
  "DASH_EMPTY=\"\"\n"
  "FILE_EXT=EXT${LEVEL_f12__does_cdb_contain_exact_match}.\0\0\0\0\0\n"
  "EXT=`echo ${FILE_EXT} | tr '[:upper:]' '[:lower:]' | sed -e \"s/\\./:/g\"`\n"
  "TEMP_FILENAME=${HOMEDIR}/.qmail${DASH}${EXT}\n"
  "VPOPMAIL_USERNAME=${EXT//:/.}${REAL_USERNAME:${#VIRTUALDOMAINS_USERNAME_ADDITION}:$[${#REAL_USERNAME}-${#VIRTUALDOMAINS_USERNAME_ADDITION}]}\n"
  "VPOPMAIL_ADDRESS=${VPOPMAIL_USERNAME}@${REAL_DOMAIN}\n"
  "echo \"=${PATTERN}:${REAL_USERNAME}:$6:$7:${HOMEDIR}:${DASH}:${FILE_EXT}:\" >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  int saved_dash_emptyable = dash_emptyable;
  int saved_dash_empty = dash_empty;
  dash_emptyable = 1;
  dash_empty = 0;
  f15__does_exact_match_home_directory_exist();
  dash_emptyable = saved_dash_emptyable;
  dash_empty = saved_dash_empty;,
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  "cdb does not contain recipient exact match",
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_exact_match} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.${LEVEL_f12__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n",
  f13__does_cdb_contain_wildcard_match();,
  "cp ${TMPDIR}/${TEST_NUM}-f12_assign.bu.${LEVEL_f12__does_cdb_contain_exact_match} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n")


DECISION(11, f11__is_virtualdomains_entry_domain_name,
  "virtualdomains entry exactly matches domain name",
  "VIRTUALDOMAINS_REAL_USERNAME[${LEVEL_f11__is_virtualdomains_entry_domain_name}]=${REAL_USERNAME}\n"
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.bu.${LEVEL_f11__is_virtualdomains_entry_domain_name}\n"
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.${LEVEL_f11__is_virtualdomains_entry_domain_name}\n"
  "echo \"${REAL_DOMAIN}:${REAL_DOMAIN}\" >> ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.${LEVEL_f11__is_virtualdomains_entry_domain_name}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.${LEVEL_f11__is_virtualdomains_entry_domain_name} /var/qmail/control/virtualdomains\n"
  "VIRTUALDOMAINS_USERNAME_ADDITION=${REAL_DOMAIN}-\n"
  "REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}${REAL_USERNAME}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n",
  num_username_hyphens++;
  virtualdomains_exact++;
  f12__does_cdb_contain_exact_match();
  virtualdomains_exact--;
  num_username_hyphens--;,
  "cp ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.bu.${LEVEL_f11__is_virtualdomains_entry_domain_name} /var/qmail/control/virtualdomains\n",
  "virtualdomains entry does not exactly match domain name",
  "VIRTUALDOMAINS_REAL_USERNAME[${LEVEL_f11__is_virtualdomains_entry_domain_name}]=${REAL_USERNAME}\n"
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.bu.${LEVEL_f11__is_virtualdomains_entry_domain_name}\n"
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.${LEVEL_f11__is_virtualdomains_entry_domain_name}\n"
  "echo \"${REAL_DOMAIN}:${REAL_DOMAIN}0\" >> ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.${LEVEL_f11__is_virtualdomains_entry_domain_name}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.${LEVEL_f11__is_virtualdomains_entry_domain_name} /var/qmail/control/virtualdomains\n"
  "VIRTUALDOMAINS_USERNAME_ADDITION=${REAL_DOMAIN}0-\n"
  "REAL_USERNAME=${VIRTUALDOMAINS_USERNAME_ADDITION}${REAL_USERNAME}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n",
  num_username_hyphens++;
  f12__does_cdb_contain_exact_match();
  num_username_hyphens--;,
  "cp ${TMPDIR}/${TEST_NUM}-f11_virtualdomains.bu.${LEVEL_f11__is_virtualdomains_entry_domain_name} /var/qmail/control/virtualdomains\n")


DECISION(10, f10__is_domain_in_virtualdomains,
  "domain is in virtualdomains",
  "",
  f11__is_virtualdomains_entry_domain_name();,
  "",
  "domain is not in virtualdomains",
  "cp /var/qmail/control/virtualdomains ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains} | sed -e \"/^${REAL_DOMAIN}:/d\" > ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.${LEVEL_f10__is_domain_in_virtualdomains}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.${LEVEL_f10__is_domain_in_virtualdomains} /var/qmail/control/virtualdomains\n",
  accept_message(ACCEPT_REMOTE);,
  "cp ${TMPDIR}/${TEST_NUM}-f10_virtualdomains.bu.${LEVEL_f10__is_domain_in_virtualdomains} /var/qmail/control/virtualdomains\n")


DECISION(09, f09__is_domain_in_locals,
  "domain is in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "echo ${TO_DOMAIN}${REAL_RELAYCLIENT} >> ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n",
  f12__does_cdb_contain_exact_match();,
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n",
  "domain is not in locals",
  "cp /var/qmail/control/locals ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals} | sed -e \"/^${REAL_DOMAIN}$/d\" > ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n",
  f10__is_domain_in_virtualdomains();,
  "cp ${TMPDIR}/${TEST_NUM}-f09_locals.bu.${LEVEL_f09__is_domain_in_locals} /var/qmail/control/locals\n")


DECISION_LIMITED(08, f08__is_domain_in_percenthack,
  1,
  "domain is in percenthack",
  "PERCENTHACK_TO_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${TO_USERNAME}\n"
  "PERCENTHACK_TO_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${TO_DOMAIN}\n"
  "PERCENTHACK_TO_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${TO_ADDRESS}\n"
  "PERCENTHACK_REAL_RELAYCLIENT[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_RELAYCLIENT}\n"
  "PERCENTHACK_REAL_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_USERNAME}\n"
  "PERCENTHACK_REAL_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_DOMAIN}\n"
  "PERCENTHACK_REAL_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_ADDRESS}\n"
  "PERCENTHACK_SAVED_ADDITION[${LEVEL_f08__is_domain_in_percenthack}]=${PERCENTHACK_USERNAME_ADDITION}\n"
  "if [ -f /var/qmail/control/percenthack ]\n"
  "then\n"
  "  cp /var/qmail/control/percenthack ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack} ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "else\n"
  "  touch ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "fi\n"
  "if [ -f /var/qmail/control/rcpthosts ]\n"
  "then\n"
  "  cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.bu.${LEVEL_f08__is_domain_in_percenthack}\n"
  "fi\n"
  "if [ -f /var/qmail/control/morercpthosts ]\n"
  "then\n"
  "  cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.bu.${LEVEL_f08__is_domain_in_percenthack}\n"
  "fi\n"
  "NEW_DOMAIN=\0\0.test\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  PERCENTHACK_USERNAME_ADDITION=${PERCENTHACK_USERNAME_ADDITION}%${NEW_DOMAIN}\n"
  "  echo ${REAL_DOMAIN} >> ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "  TO_USERNAME=${TO_USERNAME}%${NEW_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}\n"
  "  REAL_DOMAIN=${NEW_DOMAIN}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  REAL_RELAYCLIENT=\"\"\n"
  "else\n"
  "  PERCENTHACK_USERNAME_ADDITION=${PERCENTHACK_USERNAME_ADDITION}%${TO_DOMAIN}\n"
  "  REAL_DOMAIN=${TO_DOMAIN}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=${TO_USERNAME}%${TO_DOMAIN}\n"
  "  TO_DOMAIN=${NEW_DOMAIN}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "  if [ \"${IN_RCPTHOSTS}\" == \"true\" ]\n"
  "  then\n"
  "    cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    echo \"${NEW_DOMAIN}${REAL_RELAYCLIENT}\" >> ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    cp ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/rcpthosts\n"
  "  fi\n"
  "  if [ \"${IN_MORERCPTHOSTS}\" == \"true\" ]\n"
  "  then\n"
  "    cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    echo \"${NEW_DOMAIN}${REAL_RELAYCLIENT}\" >> ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.${LEVEL_f08__is_domain_in_percenthack}\n"
  "    cp ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/morercpthosts\n"
  "    /var/qmail/bin/qmail-newmrh\n"
  "  fi\n"
  "  echo ${NEW_DOMAIN}${REAL_RELAYCLIENT} >> ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack}\n"
  "  REAL_RELAYCLIENT=\"\"\n"
  "fi\n"
  "cp ${TMPDIR}/${TEST_NUM}-f08_percenthack.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/percenthack\n",
  f07__does_recipient_contain_percent();,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f08_percenthack.bu.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/percenthack\n"
  "else\n"
  "  rm -f /var/qmail/control/percenthack\n"
  "fi\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f08_rcpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/rcpthosts\n"
  "else\n"
  "  rm -f /var/qmail/control/rcpthosts\n"
  "fi\n"
  "if [ -f ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f08_morercpthosts.bu.${LEVEL_f08__is_domain_in_percenthack} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "else\n"
  "  rm -f /var/qmail/control/morercpthosts /var/qmail/control/morercpthosts.cdb\n"
  "fi\n",
  -1,
  "domain is not in percenthack",
  "PERCENTHACK_TO_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${TO_USERNAME}\n"
  "PERCENTHACK_TO_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${TO_DOMAIN}\n"
  "PERCENTHACK_TO_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${TO_ADDRESS}\n"
  "PERCENTHACK_REAL_USERNAME[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_USERNAME}\n"
  "PERCENTHACK_REAL_DOMAIN[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_DOMAIN}\n"
  "PERCENTHACK_REAL_ADDRESS[${LEVEL_f08__is_domain_in_percenthack}]=${REAL_ADDRESS}\n"
  "NEW_USERNAME=\"%\0\0\0\0\0\"\n"
  "if [ -z \"${TO_DOMAIN}\" ]\n"
  "then\n"
  "  REAL_USERNAME=${REAL_USERNAME:0:$[${#REAL_USERNAME}-${#REAL_RELAYCLIENT}]}${NEW_USERNAME}${REAL_RELAYCLIENT}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}${NEW_USERNAME}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}\n"
  "else\n"
  "  REAL_USERNAME=${REAL_USERNAME}${NEW_USERNAME}\n"
  "  REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "  TO_USERNAME=${TO_USERNAME:0:$[${#TO_USERNAME}-${#PERCENTHACK_USERNAME_ADDITION}]}${NEW_USERNAME}${PERCENTHACK_USERNAME_ADDITION}\n"
  "  TO_ADDRESS=${TO_USERNAME}@${TO_DOMAIN}\n"
  "fi\n",
  num_username_percents++;
  f09__is_domain_in_locals();
  num_username_percents--;,
  "")


DECISION(07, f07__does_recipient_contain_percent,
  "recipient contains a(nother) percent sign",
  "",
  f08__is_domain_in_percenthack();,
  "",
  "recipient does not contain a(nother) percent sign",
  "",
  f09__is_domain_in_locals();,
  "")


DECISION(06, f06__is_domain_in_morercpthosts,
  "domain is in morercpthosts",
  "IN_MORERCPTHOSTS=\"true\"\n"
  "if [ -f /var/qmail/control/morercpthosts ]\n"
  "then\n"
  "  cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "fi\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "/var/qmail/bin/qmail-newmrh\n",
  f07__does_recipient_contain_percent();,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "else\n"
  "  rm -f /var/qmail/control/morercpthosts /var/qmail/control/morercpthosts.cdb\n"
  "fi\n",
  "domain is not in morercpthosts",
  "IN_MORERCPTHOSTS=\"false\"\n"
  "if [ -f /var/qmail/control/morercpthosts ]\n"
  "then\n"
  "  cp /var/qmail/control/morercpthosts ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "  cat ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} | sed -e \"/^${REAL_DOMAIN}$/d\" > ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts}\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "fi\n",
  reject_message(REJECTION_RELAYING);,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f06_morercpthosts.bu.${LEVEL_f06__is_domain_in_morercpthosts} /var/qmail/control/morercpthosts\n"
  "  /var/qmail/bin/qmail-newmrh\n"
  "else\n"
  "  rm -f /var/qmail/control/morercpthosts /var/qmail/control/morercpthosts.cdb\n"
  "fi\n")


DECISION(05, f05__is_domain_in_rcpthosts,
  "domain is in rcpthosts",
  "IN_RCPTHOSTS=\"true\"\n"
  "cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n",
  f07__does_recipient_contain_percent();,
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n",
  "domain is not in rcpthosts",
  "IN_RCPTHOSTS=\"false\"\n"
  "cp /var/qmail/control/rcpthosts ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts} | sed -e \"/^${REAL_DOMAIN}$/d\" > ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n",
  f06__is_domain_in_morercpthosts();,
  "cp ${TMPDIR}/${TEST_NUM}-f05_rcpthosts.bu.${LEVEL_f05__is_domain_in_rcpthosts} /var/qmail/control/rcpthosts\n")


DECISION(04, f04__does_me_exist,
  "me exists",
  "ME_REAL_DOMAIN[${LEVEL_f04__does_me_exist}]=${REAL_DOMAIN}\n"
  "if [ -f /var/qmail/control/me ]\n"
  "then\n"
  "  cp /var/qmail/control/me ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist}\n"
  "fi\n"
  "REAL_DOMAIN=\0\0.test\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f04_me.${LEVEL_f04__does_me_exist}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f04_me.${LEVEL_f04__does_me_exist} /var/qmail/control/me\n",
  f07__does_recipient_contain_percent();,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} /var/qmail/control/me\n"
  "else\n"
  "  rm /var/qmail/control/me\n"
  "fi\n",
  "me does not exist",
  "ME_REAL_DOMAIN[${LEVEL_f04__does_me_exist}]=${REAL_DOMAIN}\n"
  "REAL_DOMAIN=envnoathost\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "if [ -f /var/qmail/control/me ]\n"
  "then\n"
  "  cp /var/qmail/control/me ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist}\n"
  "  rm /var/qmail/control/me\n"
  "fi\n",
  reject_message(REJECTION_CANNOT_START);,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f04_me.bu.${LEVEL_f04__does_me_exist} /var/qmail/control/me\n"
  "fi\n")


DECISION(03, f03__does_envnoathost_exist,
  "envnoathost exists",
  "if [ -f /var/qmail/control/envnoathost ]\n"
  "then\n"
  "  cp /var/qmail/control/envnoathost ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist}\n"
  "fi\n"
  "ENVNOATHOST_REAL_DOMAIN[${LEVEL_f03__does_envnoathost_exist}]=${REAL_DOMAIN}\n"
  "REAL_DOMAIN=\0\0.test\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "echo ${REAL_DOMAIN} > ${TMPDIR}/${TEST_NUM}-f03_envnoathost.${LEVEL_f03__does_envnoathost_exist}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f03_envnoathost.${LEVEL_f03__does_envnoathost_exist} /var/qmail/control/envnoathost\n",
  envnoathost_exists = 1;
  f07__does_recipient_contain_percent();,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} /var/qmail/control/envnoathost\n"
  "else\n"
  "  rm /var/qmail/control/envnoathost\n"
  "fi\n",
  "envnoathost does not exist",
  "if [ -f /var/qmail/control/envnoathost ]\n"
  "then\n"
  "  cp /var/qmail/control/envnoathost ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist}\n"
  "  rm /var/qmail/control/envnoathost\n"
  "fi\n",
  envnoathost_exists = 0;
  f04__does_me_exist();,
  "if [ -f ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} ]\n"
  "then\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f03_envnoathost.bu.${LEVEL_f03__does_envnoathost_exist} /var/qmail/control/envnoathost\n"
  "fi\n")


DECISION(02, f02__does_recipient_have_domain_name,
  "recipient has domain name",
  "",
  if (relayclient_set)
    f07__does_recipient_contain_percent();
  else
    f05__is_domain_in_rcpthosts();,
  "",
  "recipient does not have a domain name",
  "REAL_DOMAIN=\"\"\n"
  "REAL_USERNAME=${REAL_USERNAME}${REAL_RELAYCLIENT}\n"
  "REAL_ADDRESS=${REAL_USERNAME}\n"
  "TO_DOMAIN=\"\"\n"
  "TO_ADDRESS=${TO_USERNAME}\n",
  if (spamdyke)
    reject_message(REJECTION_UNQUALIFIED);
  else
    f03__does_envnoathost_exist();,
  "")


DECISION(01, f01__is_relayclient_set,
  "RELAYCLIENT is set",
  "export RELAYCLIENT=\"foo\0\0\"\n"
  "REAL_RELAYCLIENT=${RELAYCLIENT}\n"
  "REAL_DOMAIN=${REAL_DOMAIN}${REAL_RELAYCLIENT}\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n",
  relayclient_set = 1;
  f02__does_recipient_have_domain_name();
  relayclient_set = 0;,
  "export -n RELAYCLIENT\n",
  "RELAYCLIENT is not set",
  "export -n RELAYCLIENT\n",
  relayclient_set = 0;
  f02__does_recipient_have_domain_name();,
  "")


STEP(00, f00__init,
  "Initial setup",
  "outcome=\"\"\n"
  INC_DEC_FUNC("CURRENT_VALID","4")
  INC_DEC_FUNC("CURRENT_REMOTE","4")
  "export TCPREMOTEIP=11.22.33.44\n"
  "unset GREP_TARGET\n"
  "unset REJECT_GREP_TARGET\n"
  "unset TEMP_FILENAME\n"
  "unset HOMEDIR\n"
  "unset DASH\n"
  "unset EXT\n"
  "unset REAL_RELAYCLIENT\n"
  "unset REAL_USERNAME\n"
  "unset REAL_DOMAIN\n"
  "unset REAL_ADDRESS\n"
  "unset TO_USERNAME\n"
  "unset TO_DOMAIN\n"
  "unset TO_ADDRESS\n"
  "unset TEMP_USERNAME\n"
  "unset WILDCARD_DOT\n"
  "unset WILDCARD_DASH\n"
  "unset WILDCARD_DASH2\n"
  "unset STRLEN_REAL_USERNAME\n"
  "unset PATTERN\n"
  "unset WILDCARD_ADDITION\n"
  "unset IN_MORERCPTHOSTS\n"
  "unset IN_RCPTHOSTS\n"
  "unset PERCENTHACK_USERNAME_ADDITION\n"
  "unset VIRTUALDOMAINS_USERNAME_ADDITION\n"
  "TEST_SPAMDYKE_QRV_PATH=${SPAMDYKE_QRV_PATH}\n"
  "SPAMDYKE_QRV_WITH_VPOPMAIL_PATH=${SPAMDYKE_QRV_PATH}\n"
  "FROM_ADDRESS=test-${TEST_NUM}.\0\0\0\0\0.\0\0\0\0\0@example.com\n"
  "REAL_USERNAME=s-\0\0\0\n"
  "REAL_DOMAIN=\0\0.test\n"
  "REAL_ADDRESS=${REAL_USERNAME}@${REAL_DOMAIN}\n"
  "TO_USERNAME=${REAL_USERNAME}\n"
  "TO_DOMAIN=${REAL_DOMAIN}\n"
  "TO_ADDRESS=${REAL_ADDRESS}\n"
  "TEMP_USERNAME=${REAL_USERNAME}\n"
  "CURRENT_VALID=0\n"
  "CURRENT_REMOTE=0\n"
  "VALID_DOMAIN[0]=valid-domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n"
  "VALID_USERNAME[0]=valid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "INVALID_USERNAME[0]=invalid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "VALID_DOMAIN[1]=valid-domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n"
  "VALID_USERNAME[1]=valid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "INVALID_USERNAME[1]=invalid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "VALID_DOMAIN[2]=valid-domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n"
  "VALID_USERNAME[2]=valid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "INVALID_USERNAME[2]=invalid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "VALID_DOMAIN[3]=valid-domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n"
  "VALID_USERNAME[3]=valid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "INVALID_USERNAME[3]=invalid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "VALID_DOMAIN[4]=valid-domain-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0.test\n"
  "VALID_USERNAME[4]=valid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "INVALID_USERNAME[4]=invalid-user-\0\0\0\0\0-\0\0\0\0\0-\0\0\0\0\0\n"
  "cp /var/qmail/users/assign ${TMPDIR}/${TEST_NUM}-f00_assign.bu.${LEVEL_f00__init}\n"
  "cat ${TMPDIR}/${TEST_NUM}-f00_assign.bu.${LEVEL_f00__init} | sed -e \"/^=${REAL_USERNAME}:.*$/d\" -e \"/^\\.$/d\" > ${TMPDIR}/${TEST_NUM}-f00_assign.${LEVEL_f00__init}\n"
  "for i in 0 1 2 3 4\n"
  "do\n"
  "  VALID_ADDRESS[$i]=${VALID_USERNAME[$i]}@${VALID_DOMAIN[$i]}\n"
  "  INVALID_ADDRESS[$i]=${INVALID_USERNAME[$i]}@${VALID_DOMAIN[$i]}\n"
  "  VALID_HOMEDIR[$i]=${TMPDIR}/${TEST_NUM}-valid_home_${i}\n"
  "  VALID_GREP_TARGET[$i]=${VALID_HOMEDIR[$i]}/mbox\n"
  "  mkdir -p ${VALID_HOMEDIR[$i]}\n"
  "  echo ${VALID_HOMEDIR[$i]}/mbox > ${VALID_HOMEDIR[$i]}/.qmail\n"
  "  touch ${VALID_HOMEDIR[$i]}/mbox\n"
  "  sync\n"
  "  chmod 700 ${VALID_HOMEDIR[$i]}\n"
  "  chown -R $6 ${VALID_HOMEDIR[$i]}\n"
  "  chgrp -R $7 ${VALID_HOMEDIR[$i]}\n"
  "  chmod 600 ${VALID_HOMEDIR[$i]}/mbox\n"
  "  chown $6 ${VALID_HOMEDIR[$i]}/mbox\n"
  "  chgrp $7 ${VALID_HOMEDIR[$i]}/mbox\n"
  "  echo \"=${VALID_USERNAME[$i]}:${VALID_USERNAME[$i]}:$6:$7:${VALID_HOMEDIR[$i]}:::\" >> ${TMPDIR}/${TEST_NUM}-f00_assign.${LEVEL_f00__init}\n"
  "done\n"
  "echo \".\" >> ${TMPDIR}/${TEST_NUM}-f00_assign.${LEVEL_f00__init}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f00_assign.${LEVEL_f00__init} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "cp /etc/named.conf ${TMPDIR}/${TEST_NUM}-f00_named.conf.bu.${LEVEL_f00__init}\n"
  "cp /etc/named.conf ${TMPDIR}/${TEST_NUM}-f00_named.conf.${LEVEL_f00__init}\n"
  "for dom in ${VALID_DOMAIN[0]} ${VALID_DOMAIN[1]} ${VALID_DOMAIN[2]} ${VALID_DOMAIN[3]} ${VALID_DOMAIN[4]} example.com 33.22.11.in-addr.arpa\n"
  "do\n"
  "  echo \"zone \\\"${dom}\\\" { type master; file \\\"${dom}\\\"; };\" >> ${TMPDIR}/${TEST_NUM}-f00_named.conf.${LEVEL_f00__init}\n"
  "  echo \"\\$TTL 300\" > ${TMPDIR}/${TEST_NUM}-f00_named_${dom}.${LEVEL_f00__init}\n"
  "  echo \"${dom}. 300 IN SOA host.${dom}. samc.silence.org. ( 1 300 300 300 300 )\" >> ${TMPDIR}/${TEST_NUM}-f00_named_${dom}.${LEVEL_f00__init}\n"
  "  echo \"@ 300 IN NS ns.${dom}.\" >> ${TMPDIR}/${TEST_NUM}-f00_named_${dom}.${LEVEL_f00__init}\n"
  "  echo \"@ 300 IN MX 10 mail.${dom}.\" >> ${TMPDIR}/${TEST_NUM}-f00_named_${dom}.${LEVEL_f00__init}\n"
  "  echo \"ns 300 IN A 127.0.0.1\" >> ${TMPDIR}/${TEST_NUM}-f00_named_${dom}.${LEVEL_f00__init}\n"
  "done\n"
  "for i in 0 1 2 3 4\n"
  "do\n"
  "  echo \"mail 300 IN A 127.0.0.1\" >> ${TMPDIR}/${TEST_NUM}-f00_named_${VALID_DOMAIN[$i]}.${LEVEL_f00__init}\n"
  "done\n"
  "echo \"mail 300 IN A 172.31.254.254\" >> ${TMPDIR}/${TEST_NUM}-f00_named_example.com.${LEVEL_f00__init}\n"
  "echo \"44 300 IN PTR host.${VALID_DOMAIN[0]}.\" >> ${TMPDIR}/${TEST_NUM}-f00_named_33.22.11.in-addr.arpa.${LEVEL_f00__init}\n"
  "cp ${TMPDIR}/${TEST_NUM}-f00_named.conf.${LEVEL_f00__init} /etc/named.conf\n"
  "for dom in ${VALID_DOMAIN[0]} ${VALID_DOMAIN[1]} ${VALID_DOMAIN[2]} ${VALID_DOMAIN[3]} ${VALID_DOMAIN[4]} example.com 33.22.11.in-addr.arpa\n"
  "do\n"
  "  cp ${TMPDIR}/${TEST_NUM}-f00_named_${dom}.${LEVEL_f00__init} /var/named/${dom}\n"
  "done\n"
  "if [ -x /bin/systemctl ]\n"
  "then\n"
  "  /bin/systemctl reload named\n"
  "else\n"
  "  /etc/rc.d/init.d/named reload\n"
  "fi\n",
  f01__is_relayclient_set();,
  "cp ${TMPDIR}/${TEST_NUM}-f00_assign.bu.${LEVEL_f00__init} /var/qmail/users/assign\n"
  "sync\n"
  "/var/qmail/bin/qmail-newu\n"
  "cp ${TMPDIR}/${TEST_NUM}-f00_named.conf.bu.${LEVEL_f00__init} /etc/named.conf\n"
  "for dom in ${VALID_DOMAIN[0]} ${VALID_DOMAIN[1]} ${VALID_DOMAIN[2]} ${VALID_DOMAIN[3]} ${VALID_DOMAIN[4]} example.com 33.22.11.in-addr.arpa\n"
  "do\n"
  "  rm -f /var/named/${dom}\n"
  "done\n"
  "if [ -x /bin/systemctl ]\n"
  "then\n"
  "  /bin/systemctl reload named\n"
  "else\n"
  "  /etc/rc.d/init.d/named reload\n"
  "fi\n"
  )

void usage()
  {
  fprintf(stderr,
    "USAGE: generator [ -v ] [ -b NUM ] [ -e NUM ] [ -p PATTERN ] [ -q ] [ -s ] target_dir\n"
    "\n"
    "Within target_dir, generator creates the test scripts necessary to completely\n"
    "test qmail's recipient validation procedure.  Two versions of each possible test\n"
    "can be created: one that tests the scenario with qmail only and one that tests\n"
    "the scenario with spamdyke.\n"
    "\n"
    "Generating the full list of tests will consume gigabytes of disk space.  Using\n"
    "the -b and -e options to limit the number of scripts is advisable.\n"
    "\n"
    "target_dir must be a directory and no test scripts must exist within it;\n"
    "generator will not overwrite any files.\n"
    "\n"
    "-b NUM\n"
    "  Start generating scripts at number NUM\n"
    "\n"
    "-e NUM\n"
    "  Stop generating scripts at number NUM\n"
    "\n"
    "-p PATTERN\n"
    "  Only generate scripts whose sequence of XXyXXn characters contains PATTERN\n"
    "\n"
    "-q\n"
    "  Create the scripts to test with qmail only\n"
    "\n"
    "-s\n"
    "  Create the scripts to test with spamdyke\n"
    "\n"
    "-v\n"
    "  Produce verbose output during processing (may be given multiple times)\n"
    "\n"
    "Either -q or -s or both must be given.\n"
    );

  exit(1);

  return;
  }

void reinit()
  {
  strlen_comments = 0;
  comments[0] = '\0';
  strlen_pre = 0;
  pre[0] = '\0';
  strlen_post = 0;
  post[0] = '\0';
  strlen_path = 0;
  path[0] = '\0';
  test_name[0] = '\0';

  relayclient_set = 0;
  using_defaultdelivery = 0;
  dash_emptyable = 1;
  dash_empty = 1;
  num_ext_hyphens = 0;
  num_username_hyphens = 1;
  num_username_percents = 0;
  forwards_only = 0;
  lines_in_file = 0;
  comments_in_file = 0;
  username_truncated = 0;
  invalid_forwards = 0;
  virtualdomains_exact = 0;
  assign_wildcard_domain = 0;
  spamdyke_blind = 0;
  vdelivermail_loop = 0;

  return;
  }

int main(int argc, char *argv[])
  {
  int i;
  int do_qmail;
  int do_spamdyke;
  int opt;
  int tmp_int;

  opterr = 0;
  do_qmail = 0;
  do_spamdyke = 0;
  total_number = 0;
  next_number = 0;
  start_time = 0;
  max_strlen = 0;
  max_path = 0;
  test_dir[0] = '\0';
  begin_num = -1;
  end_num = -1;
  strlen_target_pattern = 0;
  verbose = 0;

  while ((opt = getopt(argc, argv, "d:qsb:e:p:v")) != -1)
    {
    switch (opt)
      {
      case 'q':
        do_qmail = 1;
        break;
      case 's':
        do_spamdyke = 1;
        break;
      case 'b':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          begin_num = tmp_int;

        break;
      case 'e':
        if ((sscanf(optarg, "%d", &tmp_int) == 1) &&
            (tmp_int >= 0))
          end_num = tmp_int;

        break;
      case 'p':
        strlen_target_pattern = MINVAL(MAX_BUF, strlen(optarg));
        memcpy(target_pattern, optarg, strlen_target_pattern);
        target_pattern[strlen_target_pattern] = '\0';
        break;
      case 'v':
        verbose = 1;
        break;
      default:
        usage();
        break;
      }
    }

  if ((argv[optind] != NULL) &&
      (do_qmail ||
       do_spamdyke))
    {
    srandom(time(NULL));

    memcpy(test_dir, argv[optind], MINVAL(strlen(argv[optind]), MAX_BUF));
    for (i = 0; test_dir[i] != '\0'; i++);
    for (i--; test_dir[i] == '/'; i--)
      test_dir[i] = '\0';

    reinit();
    count_only = 1;

    if (do_qmail)
      {
      spamdyke = 0;
      f00__init();
      }

    reinit();

    if (do_spamdyke)
      {
      spamdyke = 1;
      f00__init();
      }

    total_number--;
    count_only = 0;

    if (MAX_BUF < max_strlen)
      {
      printf("ERROR: Largest buffer needed is %d bytes, allocation is set to %d\nIncrease MAX_BUF to fix this problem.\n", max_strlen, MAX_BUF);
      exit(100);
      }
    if (num_username_hyphens != 1)
      {
      printf("ERROR: num_username_hyphens is %d, should be 1!", num_username_hyphens);
      exit(100);
      }

    if ((max_path + MAX_SUFFIX + (int)ceil(log10(total_number)) + 1) > MAX_FOLDER_LEN)
      {
      printf("ERROR: longest path to be generated is %d, maximum allowed in filesystem is %d!", max_path + MAX_SUFFIX + (int)ceil(log10(total_number)) + 1, MAX_FOLDER_LEN);
      exit(100);
      }

    reinit();
    start_time = time(NULL);

    if (do_qmail)
      {
      spamdyke = 0;
      f00__init();
      }

    reinit();

    if (do_spamdyke)
      {
      spamdyke = 1;
      f00__init();
      }
    }
  else
    usage();

  return(0);
  }
